import itertools

import numpy as np
import pytest

import pyxu.info.deps as pxd
import pyxu.operator as pxo
import pyxu.runtime as pxrt
import pyxu_tests.operator.conftest as conftest


class SubSampleMixin:
    @pytest.fixture
    def arg_shape(self, request):  # anything which is understood as shape parameter of np.zeros()
        raise NotImplementedError

    @pytest.fixture
    def indices(self, request):  # anything which can index an NDArray
        raise NotImplementedError

    @pytest.fixture
    def sub_shape(self, arg_shape, indices) -> tuple[int]:
        sh = np.zeros(arg_shape)[indices].shape
        return sh

    @pytest.fixture
    def data_shape(self, arg_shape, sub_shape):
        return (np.prod(sub_shape), np.prod(arg_shape))

    @pytest.fixture
    def data_apply(self, arg_shape, indices):
        arr = np.arange(np.prod(arg_shape)).reshape(arg_shape)
        out = arr[indices].reshape(-1)
        return dict(
            in_=dict(arr=arr.reshape(-1)),
            out=out,
        )


class TestSubSample(SubSampleMixin, conftest.LinOpT):
    @pytest.fixture(
        params=[
            10,
            (10,),
            (10, 5, 2),
        ]
    )
    def arg_shape(self, request):
        return request.param

    @pytest.fixture(
        params=[
            0,  # integer indexing
            [1],  # list indexing
            slice(0, None, 2),
            np.r_[1],  # using an NDArray
            [i % 2 == 0 for i in range(10)],  # boolean mask
        ]
    )
    def indices(self, request):
        return (request.param,)

    @pytest.fixture(
        params=itertools.product(
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def spec(self, arg_shape, indices, request):
        ndi, width = request.param
        op = pxo.SubSample(arg_shape, *indices)
        return op, ndi, width


class TestTrim(SubSampleMixin, conftest.LinOpT):
    @pytest.fixture(
        params=[
            (
                (10, 5, 11),
                0,  # no trimming
                (slice(None), slice(None), slice(None)),
            ),
            (
                (10, 5, 11),
                1,  # equal trim/dimension
                (slice(1, 9), slice(1, 4), slice(1, 10)),
            ),
            (
                (10, 5, 11),
                (1, 2, 3),  # different trim/dimension
                (slice(1, 9), slice(2, 3), slice(3, 8)),
            ),
            (
                (10, 5, 11),
                ((0, 0), (1, 0), (2, 3)),  # different head/tail/dimension
                (slice(None), slice(1, 5), slice(2, 8)),
            ),
        ]
    )
    def _spec(self, request):
        # (arg_shape, trim_width, index_spec-equivalent) configs to test
        arg_shape, trim_width, index_spec = request.param
        return arg_shape, trim_width, index_spec

    @pytest.fixture
    def arg_shape(self, _spec):
        arg_shape, _, _ = _spec
        return arg_shape

    @pytest.fixture
    def trim_width(self, _spec):
        _, trim_width, _ = _spec
        return trim_width

    @pytest.fixture
    def indices(self, _spec):
        _, _, index = _spec
        return index

    @pytest.fixture(
        params=itertools.product(
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def spec(self, arg_shape, trim_width, request):
        ndi, width = request.param
        op = pxo.Trim(arg_shape, trim_width)
        return op, ndi, width



import itertools

import numpy as np
import pytest
import scipy.ndimage as snd

import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.operator as pxo
import pyxu.runtime as pxrt
import pyxu_tests.operator.conftest as conftest


# We disable PrecisionWarnings since Stencil() is not precision-agnostic.
# We disable NumbaPerformanceWarnings due to solving small-scale problems at test time.
@pytest.mark.filterwarnings("ignore::pyxu.info.warning.PrecisionWarning")
@pytest.mark.filterwarnings("ignore::numba.core.errors.NumbaPerformanceWarning")
class TestStencil(conftest.SquareOpT):
    @pytest.fixture(
        params=[
            # 1D, random center/mode ------------------------
            (
                (10, np.arange(1, 7), (0,), "constant"),
                ((10,), (np.arange(1, 7),), ((0,),), ("constant",)),
            ),
            (
                (10, np.arange(1, 7), (1,), "edge"),
                ((10,), (np.arange(1, 7),), ((1,),), ("edge",)),
            ),
            (
                (10, np.arange(1, 7), (2,), "wrap"),
                ((10,), (np.arange(1, 7),), ((2,),), ("wrap",)),
            ),
            (
                (10, np.arange(1, 7), (3,), "reflect"),
                ((10,), (np.arange(1, 7),), ((3,),), ("reflect",)),
            ),
            (
                (10, np.arange(1, 7), (4,), "symmetric"),
                ((10,), (np.arange(1, 7),), ((4,),), ("symmetric",)),
            ),
            # ND, random center/mode ------------------------
            (
                ((10, 11), np.arange(1, 9).reshape(2, 4), (0, 3), "constant"),
                ((10, 11), (np.arange(1, 9).reshape(2, 4),), ((0, 3),), ("constant", "constant")),
            ),
            (
                ((10, 11), np.arange(1, 9).reshape(2, 4), (1, 2), ("wrap", "reflect")),
                ((10, 11), (np.arange(1, 9).reshape(2, 4),), ((1, 2),), ("wrap", "reflect")),
            ),
            (
                ((10, 11), np.arange(1, 9).reshape(2, 4), (1, 1), ("edge", "symmetric")),
                ((10, 11), (np.arange(1, 9).reshape(2, 4),), ((1, 1),), ("edge", "symmetric")),
            ),
            # ND seperable, random center/mode --------------
            (
                ((10, 11), (np.arange(1, 7), np.arange(2, 5)), (3, 0), "constant"),
                (
                    (10, 11),
                    (np.arange(1, 7).reshape(-1, 1), np.arange(2, 5).reshape(1, -1)),
                    ((3, 0), (0, 0)),
                    ("constant", "constant"),
                ),
            ),
            (
                ((10, 11), (np.arange(1, 7), np.arange(2, 5)), (2, 1), ("edge", "wrap")),
                (
                    (10, 11),
                    (np.arange(1, 7).reshape(-1, 1), np.arange(2, 5).reshape(1, -1)),
                    ((2, 0), (0, 1)),
                    ("edge", "wrap"),
                ),
            ),
            (
                ((10, 11), (np.arange(1, 7), np.arange(2, 5)), (3, 2), ("reflect", "symmetric")),
                (
                    (10, 11),
                    (np.arange(1, 7).reshape(-1, 1), np.arange(2, 5).reshape(1, -1)),
                    ((3, 0), (0, 2)),
                    ("reflect", "symmetric"),
                ),
            ),
        ]
    )
    def _spec(self, request):
        # (arg_shape, kernel, center, mode) configs to test
        # * `request.param[0]` corresponds to raw inputs users provide to Stencil().
        # * `request.param[1]` corresponds to their ground-truth canonical parameterization.
        return request.param

    @pytest.fixture
    def arg_shape(self, _spec):  # canonical representation
        arg_shape, _, _, _ = _spec[1]
        return arg_shape

    @pytest.fixture
    def kernel(self, _spec):  # canonical representation (NumPy)
        _, kernel, _, _ = _spec[1]
        return kernel

    @pytest.fixture
    def center(self, _spec):  # canonical representation
        _, _, center, _ = _spec[1]
        return center

    @pytest.fixture
    def mode(self, _spec):  # canonical representation
        _, _, _, mode = _spec[1]
        return mode

    @pytest.fixture(
        params=itertools.product(
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def spec(self, _spec, request) -> tuple[pxt.OpT, pxd.NDArrayInfo, pxrt.Width]:
        ndi, width = request.param
        arg_shape, kernel, center, mode = _spec[0]  # user-provided form

        # transform kernel to right backend
        self._skip_if_unsupported(ndi)
        xp = ndi.module()
        try:
            pxd.NDArrayInfo.from_obj(kernel)  # passes if array object
            kernel = xp.array(kernel)
        except Exception:
            kernel = [xp.array(k) for k in kernel]

        with pxrt.Precision(width):
            op = pxo.Stencil(
                arg_shape=arg_shape,
                kernel=kernel,
                center=center,
                mode=mode,
            )
        return op, ndi, width

    @pytest.fixture
    def data_shape(self, arg_shape) -> pxt.OpShape:
        codim = dim = np.prod(arg_shape)
        return (codim, dim)

    @pytest.fixture(params=[0, 1, 2])  # different seeds to test robustness
    def data_apply(
        self,
        arg_shape,
        kernel,
        center,
        mode,
        request,
    ) -> conftest.DataLike:
        arr = self._random_array(arg_shape, seed=request.param)

        # Pad input in excess of what is stricly required (using Pad(); assumed correct)
        if len(kernel) == 1:  # non-seperable filter
            pad_width = [(w, w) for w in kernel[0].shape]
        else:  # seperable filter(s)
            pad_width = [(k.shape[i],) * 2 for (i, k) in enumerate(kernel)]
        pad = pxo.Pad(
            arg_shape=arg_shape,
            pad_width=pad_width,
            mode=mode,
        )
        corr_in = pad.apply(arr.reshape(-1)).reshape(pad._pad_shape)

        # perform correlation via scipy.ndimage.correlate
        corr_out = corr_in.copy()
        for k, c in zip(kernel, center):
            origin = [cc - (n // 2) for (cc, n) in zip(c, k.shape)]
            corr_out = snd.correlate(
                input=corr_out,
                weights=k,
                mode="constant",
                cval=0,
                origin=origin,
            )

        # Trim fat off (using Trim(); assumed correct)
        trim = pxo.Trim(
            arg_shape=corr_out.shape,
            trim_width=pad_width,
        )
        out = trim.apply(corr_out.reshape(-1)).reshape(arg_shape)

        return dict(
            in_=dict(arr=arr.reshape(-1)),
            out=out.reshape(-1),
        )


class TestConvolve:
    # A convolution corresponds to a stencil with reversed kernel/center.
    # Since Convolve() inherits from Stencil(), it is sufficient to test that convolution with
    # shifted Diracs (in 1D) gives the right results.
    @pytest.mark.parametrize(
        ["kernel", "center", "shift"],
        [
            (np.r_[1, 0], (0,), 0),  # k[n] = \delta[n]
            (np.r_[0, 1], (0,), -1),  # k[n] = \delta[n - 1]
            (np.r_[0, 0, 1, 0], (1,), -1),  # k[n] = \delta[n - 1]
            (np.r_[1, 0], (1,), 1),  # k[n] = \delta[n + 1]
            (np.r_[1, 0, 0], (1,), 1),  # k[n] = \delta[n + 1]
            (np.r_[0, 0, 1, 0, 0, 0], (3,), 1),  # k[n] = \delta[n + 1]
        ],
    )
    def test_value1D_apply(self, kernel, center, shift):
        N = 10
        arr = np.arange(N)

        # Compute ground-truth
        if shift == 0:
            out_gt = arr.copy()
        elif shift > 0:  # anti-causal filter
            out_gt = np.pad(arr, pad_width=(0, abs(shift)))[-N:]
        else:  # causal filter
            out_gt = np.pad(arr, pad_width=(abs(shift), 0))[:N]

        # Compute Stencil-based solution
        op = pxo.Convolve(
            arg_shape=arr.shape,
            kernel=kernel,
            center=center,
            mode="constant",
        )
        out = op.apply(arr)

        assert TestStencil._metric(out, out_gt, as_dtype=out.dtype)



import numpy as np
import pytest

import pyxu.info.ptype as pxt
import pyxu.util as pxu
import pyxu_tests.operator.conftest as conftest


class NUFFT_Mixin:
    # Internal helpers --------------------------------------------------------
    @classmethod
    def _metric(
        cls,
        a: pxt.NDArray,
        b: pxt.NDArray,
        as_dtype: pxt.DType,
    ) -> bool:
        # NUFFT is an approximate transform.
        # Based on [FINUFFT], results hold up to a small relative error.
        #
        # We choose a conservative threshold, irrespective of the `eps` parameter chosen by the
        # user. Additional tests below test explicitly if computed values correctly obey `eps`.
        eps_default = 1e-2

        cast = lambda x: pxu.compute(x)
        lhs = np.linalg.norm(pxu.to_NUMPY(cast(a) - cast(b)), axis=-1)
        rhs = np.linalg.norm(pxu.to_NUMPY(cast(b)), axis=-1)
        return conftest.less_equal(lhs, eps_default * rhs, as_dtype=as_dtype).all()

    # Fixtures which parametrize operator -------------------------------------
    @pytest.fixture(params=[1, 2, 3])
    def transform_dimension(self, request) -> int:
        return request.param

    @pytest.fixture(params=[-1, 1])
    def transform_sign(self, request) -> int:
        return request.param

    @pytest.fixture(params=[0, 1e-4, 1e-6])
    def transform_eps(self, request) -> float:
        return request.param

    @pytest.fixture(params=[1, 2])
    def transform_ntrans(self, request) -> int:
        return request.param

    @pytest.fixture(params=[1, 2])
    def transform_nthreads(self, request) -> int:
        return request.param

    @pytest.fixture(params=[False, True])
    def transform_real(self, request) -> bool:
        return request.param

    @pytest.fixture(
        params=[
            0,  # LINEAR order
            1,  # FFT order
        ]
    )
    def transform_modeord(self, request) -> int:
        return request.param

    # Overridden Tests --------------------------------------------------------
    def test_value_to_sciop(self, _op_sciop, _data_to_sciop):
        if _data_to_sciop["mode"] in {"matmat", "rmatmat"}:
            pytest.xfail(reason="Last axis is non-contiguous: apply/adjoint will fail.")
        else:
            super().test_value_to_sciop(_op_sciop, _data_to_sciop)

    def test_backend_to_sciop(self, _op_sciop, _data_to_sciop):
        if _data_to_sciop["mode"] in {"matmat", "rmatmat"}:
            pytest.xfail(reason="Last axis is non-contiguous: apply/adjoint will fail.")
        else:
            super().test_backend_to_sciop(_op_sciop, _data_to_sciop)

    def test_prec_to_sciop(self, _op_sciop, _data_to_sciop):
        if _data_to_sciop["mode"] in {"matmat", "rmatmat"}:
            pytest.xfail(reason="Last axis is non-contiguous: apply/adjoint will fail.")
        else:
            super().test_prec_to_sciop(_op_sciop, _data_to_sciop)



import itertools

import numpy as np
import numpy.random as npr
import pytest

import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.operator as pxo
import pyxu.runtime as pxrt
import pyxu.util as pxu
import pyxu_tests.operator.conftest as conftest


class TestKLDivergence(conftest.ProxFuncT):
    @pytest.fixture(
        params=itertools.product(
            # dim
            (5,),
            list(pxd.NDArrayInfo),
            pxrt.Width,
        )
    )
    def _spec(self, request):
        dim = request.param[0]
        ndi = request.param[1]
        dtype = request.param[2].value.name
        data = np.arange(1, dim + 1, dtype=dtype)
        if ndi is None:
            pytest.skip(f"{ndi} unsupported on this machine.")
            op = None
        else:
            xp = ndi.module()
            op = pxo.KLDivergence(dim, xp.array(data))
        return (dim, op), ndi, request.param[2], data

    @pytest.fixture
    def spec(self, _spec):
        return _spec[0][1], _spec[1], _spec[2]

    @pytest.fixture
    def dim(self, _spec):
        return _spec[0][0]

    @pytest.fixture
    def kl_data(self, _spec):
        return _spec[3]

    @pytest.fixture
    def data_shape(self, dim):
        return (1, dim)

    @pytest.fixture(params=[np.zeros((5,)) + 1e-6, np.arange(1, 6)])  # 2 evaluation points
    def data_apply(self, request, kl_data):
        x = request.param
        y = pxu.to_NUMPY(kl_data)

        H = np.ones_like(x) * np.inf
        ids_1 = (x > 0) * (y > 0)
        ids_2 = (x == 0) * (y > 0)
        H[ids_1] = y[ids_1] * np.log(y[ids_1] / x[ids_1])
        H[ids_2] = 0
        out = np.sum(H - y + x)
        return dict(
            in_=dict(arr=x),
            out=out,
        )

    @pytest.fixture(
        params=[  # 2 evaluation points
            dict(
                arr=np.zeros((5,)) + 1e-6,
                tau=1,
            ),
            dict(
                arr=np.arange(1, 6),
                tau=1,
            ),
        ]
    )
    def data_prox(self, request, kl_data):
        # ground truth prox taken from Pyxu v1.
        in_ = request.param
        arr = in_["arr"]
        tau = in_["tau"]
        out = (arr - tau + np.sqrt((arr - tau) ** 2 + 4 * tau * kl_data)) / 2
        return dict(in_=in_, out=out)

    @staticmethod
    def _random_array(
        shape: pxt.NDArrayShape,
        seed: int = 0,
        xp: pxt.ArrayModule = pxd.NDArrayInfo.NUMPY.module(),
        width: pxrt.Width = pxrt.Width.DOUBLE,
    ):
        # Create only arrays with only positive-valued entries.
        rng = npr.default_rng(seed)
        x = rng.normal(size=shape)
        return xp.abs(xp.array(x, dtype=width.value)) + 1e-6

    @pytest.fixture
    def data_math_lipschitz(self, dim):
        N_test, dim = 10, self._sanitize(dim, 3)
        return self._random_array((N_test, dim))

    @pytest.mark.skip("KLDiv is already a loss.")
    def test_interface_asloss(self, op, xp, width):
        pass



import math

import pyxu.abc as pxa
import pyxu.info.ptype as pxt
import pyxu.info.warning as pxw
import pyxu.runtime as pxrt
import pyxu.util as pxu

__all__ = [
    "Adam",
]


class Adam(pxa.Solver):
    r"""
    Adam solver [ProxAdam]_.

    Adam minimizes

    .. math::

       {\min_{\mathbf{x}\in\mathbb{R}^N} \;\mathcal{F}(\mathbf{x})},

    where:

    * :math:`\mathcal{F}:\mathbb{R}^N\rightarrow \mathbb{R}` is *convex* and *differentiable*, with
      :math:`\beta`-*Lipschitz continuous* gradient, for some :math:`\beta\in[0,+\infty[`.

    Adam is a suitable alternative to Proximal Gradient Descent (:py:class:`~pyxu.opt.solver.PGD`) when:

    * the cost function is differentiable,
    * computing :math:`\beta` to optimally choose the step size is infeasible,
    * line-search methods to estimate step sizes are too expensive.

    Compared to PGD, Adam auto-tunes gradient updates based on stochastic estimates of :math:`\phi_{t} =
    \mathbb{E}[\nabla\mathcal{F}]` and :math:`\psi_{t} = \mathbb{E}[\nabla\mathcal{F}^{2}]` respectively.

    Adam has many named variants for particular choices of :math:`\phi` and :math:`\psi`:

    * Adam:

      .. math::

         \phi_t
         =
         \frac{
           \mathbf{m}_t
         }{
           1-\beta_1^t
         }
         \qquad
         \psi_t
         =
         \sqrt{
           \frac{
             \mathbf{v}_t
           }{
             1-\beta_2^t
           }
         } + \epsilon,

    * AMSGrad:

      .. math::

         \phi_t = \mathbf{m}_t
         \qquad
         \psi_t = \sqrt{\hat{\mathbf{v}}_t},

    * PAdam:

      .. math::

         \phi_t = \mathbf{m}_t
         \qquad
         \psi_t = \hat{\mathbf{v}}_t^p,

    where in all cases:

    .. math::

       \mathbf{m}_t
       =
       \beta_1\mathbf{m}_{t-1}
       +
       (1-\beta_1)\mathbf{g}_t \\
       \mathbf{v}_t
       =
       \beta_2\mathbf{v}_{t-1}
       +
       (1-\beta_2)\mathbf{g}_t^2\\
       \hat{\mathbf{v}}_t
       =
       \max(\hat{\mathbf{v}}_{t-1}, \mathbf{v}_t),

    with :math:`\mathbf{m}_0 = \mathbf{v}_0 = \mathbf{0}`.

    Remarks
    -------
    * The convergence is guaranteed for step sizes :math:`\alpha\leq 2/\beta`.

    * The default stopping criterion is the relative norm change of the primal variable.  By default, the algorithm
      stops when the norm of the difference between two consecutive iterates :math:`\{\mathbf{x}_n\}_{n\in\mathbb{N}}`
      is smaller than 1e-4.  Different stopping criteria can be used.  It is recommended to change the stopping
      criterion when using the PAdam and AMSGrad variants to avoid premature stops.

    Parameters (``__init__()``)
    ---------------------------
    * **f** (:py:class:`~pyxu.abc.DiffFunc`)
      --
      Differentiable function :math:`\mathcal{F}`.
    * **\*\*kwargs** (:py:class:`~collections.abc.Mapping`)
      --
      Other keyword parameters passed on to :py:meth:`pyxu.abc.Solver.__init__`.

    Parameters (``fit()``)
    ----------------------
    * **x0** (:py:attr:`~pyxu.info.ptype.NDArray`)
      --
      (..., N) initial point(s).
    * **variant** ("adam", "amsgrad", "padam")
      --
      Name of the Adam variant to use.  Defaults to "adam".
    * **a** (:py:attr:`~pyxu.info.ptype.Real`, :py:obj:`None`)
      --
      Max normalized gradient step size.  Defaults to :math:`1 / \beta` if unspecified.
    * **b1** (:py:attr:`~pyxu.info.ptype.Real`)
      --
      1st-order gradient exponential decay :math:`\beta_{1} \in [0, 1)`.
    * **b2** (:py:attr:`~pyxu.info.ptype.Real`)
      --
      2nd-order gradient exponential decay :math:`\beta_{2} \in [0, 1)`.
    * **m0** (:py:attr:`~pyxu.info.ptype.NDArray`, :py:obj:`None`)
      --
      (..., N) initial 1st-order gradient estimate corresponding to each initial point.  Defaults to the null vector if
      unspecified.
    * **v0** (:py:attr:`~pyxu.info.ptype.NDArray`, :py:obj:`None`)
      --
      (..., N) initial 2nd-order gradient estimate corresponding to each initial point.  Defaults to the null vector if
      unspecified.
    * **p** (:py:attr:`~pyxu.info.ptype.Real`)
      --
      PAdam power parameter :math:`p \in (0, 0.5]`.  Must be specified for PAdam, unused otherwise.
    * **eps_adam** (:py:attr:`~pyxu.info.ptype.Real`)
      --
      Adam noise parameter :math:`\epsilon`.  This term is used exclusively if `variant="adam"`.  Defaults to 1e-6.
    * **eps_var** (:py:attr:`~pyxu.info.ptype.Real`)
      --
      Avoids division by zero if estimated gradient variance is too small.  Defaults to 1e-6.
    * **\*\*kwargs** (:py:class:`~collections.abc.Mapping`)
      --
      Other keyword parameters passed on to :py:meth:`pyxu.abc.Solver.fit`.

    Note
    ----
    If provided, `m0` and `v0` must be broadcastable with `x0`.

    Example
    -------
    Consider the following optimization problem:

    .. math::

       \min_{\mathbf{x}\in\mathbb{R}^N} \Vert{\mathbf{x}-\mathbf{1}}\Vert_2^2

    .. code-block:: python3

       import numpy as np

       from pyxu.operator import SquaredL2Norm
       from pyxu.opt.solver import Adam

       N = 3
       f = SquaredL2Norm(dim=N).asloss(1)

       slvr = Adam(f)
       slvr.fit(
           x0=np.zeros((N,)),
           variant="padam",
           p=0.25,
       )
       x_opt = slvr.solution()
       np.allclose(x_opt, 1, rtol=1e-4)  # True
    """

    def __init__(
        self,
        f: pxa.DiffFunc,
        **kwargs,
    ):
        kwargs.update(
            log_var=kwargs.get("log_var", ("x",)),
        )
        super().__init__(**kwargs)

        self._f = f

    @pxrt.enforce_precision(i=("x0", "a", "b1", "b2", "m0", "v0", "p", "eps_adam", "eps_var"))
    def m_init(  # default values from https://github.com/pmelchior/proxmin/blob/master/proxmin/algorithms.py
        self,
        x0: pxt.NDArray,
        variant: str = "adam",
        a: pxt.Real = None,
        b1: pxt.Real = 0.9,
        b2: pxt.Real = 0.999,
        m0: pxt.NDArray = None,
        v0: pxt.NDArray = None,
        p: pxt.Real = 0.5,
        eps_adam: pxt.Real = 1e-6,
        eps_var: pxt.Real = 1e-6,
    ):
        mst = self._mstate  # shorthand
        xp = pxu.get_array_module(x0)
        mst["x"] = x0

        mst["variant"] = self.__parse_variant(variant)

        if a is None:
            g = lambda _: math.isclose(self._f.diff_lipschitz, _)
            if g(0) or g(math.inf):
                error_msg = "Cannot auto-infer step size: choose it manually."
                raise pxw.AutoInferenceWarning(error_msg)
            else:
                mst["a"] = pxrt.coerce(1 / self._f.diff_lipschitz)
        else:
            assert a > 0, f"Parameter[a] must be positive, got {a}."
            mst["a"] = a

        assert 0 <= b1 < 1, f"Parameter[b1]: expected value in [0, 1), got {b1}."
        mst["b1"] = b1

        assert 0 <= b2 < 1, f"Parameter[b2]: expected value in [0, 1), got {b2}."
        mst["b2"] = b2

        if m0 is None:
            mst["mean"] = xp.zeros_like(x0)
        elif m0.shape == x0.shape:
            # No broadcasting involved
            mst["mean"] = m0
        else:
            x0, m0 = xp.broadcast_arrays(x0, m0)
            mst["mean"] = m0.copy()

        if v0 is None:
            mst["variance"] = xp.zeros_like(x0)
        elif v0.shape == x0.shape:
            # No broadcasting involved
            mst["variance"] = v0
        else:
            x0, v0 = xp.broadcast_arrays(x0, v0)
            mst["variance"] = v0.copy()
        mst["variance_hat"] = mst["variance"]

        assert 0 < p <= 0.5, f"Parameter[p]: expected value in (0, 0.5], got {p}."
        mst["padam_p"] = p

        assert eps_adam > 0, f"Parameter[eps_adam]: expected positive value, got {eps_adam}."
        mst["eps_adam"] = eps_adam

        assert eps_var > 0, f"Parameter[eps_var]: expected positive value, got {eps_var}."
        mst["eps_variance"] = eps_var

    def m_step(self):
        mst = self._mstate  # shorthand

        x, a = mst["x"], mst["a"]
        xp = pxu.get_array_module(x)
        gm = pxu.copy_if_unsafe(self._f.grad(x))
        gv = gm.copy()

        ## Part 1: evaluate phi/psi ============================
        m, b1 = mst["mean"], mst["b1"]
        # In-place implementation of -----------------
        #   m = b1 * m + (1 - b1) * g
        m *= b1
        gm *= 1 - b1
        m += gm
        # --------------------------------------------
        mst["mean"] = m

        v, b2 = mst["variance"], mst["b2"]
        # In-place implementation of -----------------
        #   v = b2 * v + (1 - b2) * (g ** 2)
        v *= b2
        gv **= 2
        gv *= 1 - b2
        v += gv
        # --------------------------------------------
        mst["variance"] = v.clip(mst["eps_variance"], None)  # avoid division-by-zero
        mst["variance_hat"] = xp.maximum(mst["variance_hat"], mst["variance"])

        phi = self.__phi(t=self._astate["idx"])
        psi = self.__psi(t=self._astate["idx"])
        ## =====================================================

        ## Part 2: take a step in the gradient's direction =====
        # In-place implementation of -----------------
        #   x = x - a * (phi / psi)
        phi /= psi
        phi *= a
        x -= phi
        ## =====================================================

        mst["x"] = x

    def default_stop_crit(self) -> pxa.StoppingCriterion:
        from pyxu.opt.stop import RelError

        # Described in [ProxAdam]_ and used in their implementation:
        # https://github.com/pmelchior/proxmin/blob/master/proxmin/algorithms.py
        rel_error = RelError(
            eps=1e-4,
            var="x",
            f=None,
            norm=2,
            satisfy_all=True,
        )
        return rel_error

    def objective_func(self) -> pxt.NDArray:
        func = lambda x: self._f.apply(x)
        y = func(self._mstate["x"])
        return y

    def solution(self) -> pxt.NDArray:
        """
        Returns
        -------
        x: NDArray
            (..., N) solution.
        """
        data, _ = self.stats()
        return data.get("x")

    def __phi(self, t: int):
        mst = self._mstate
        var = mst["variant"]
        if var == "adam":
            out = mst["mean"].copy()
            out /= 1 - (mst["b1"] ** t)
        elif var in ["amsgrad", "padam"]:
            out = mst["mean"].copy()  # to allow in-place updates outside __compute_phi()
        return out

    def __psi(self, t: int):
        mst = self._mstate
        xp = pxu.get_array_module(mst["x"])
        var = mst["variant"]
        if var == "adam":
            out = xp.sqrt(mst["variance"])
            out /= xp.sqrt(1 - mst["b2"] ** t)
            out += mst["eps_adam"]
        elif var == "amsgrad":
            out = xp.sqrt(mst["variance_hat"])
        elif var == "padam":
            out = mst["variance_hat"] ** mst["padam_p"]
        return out

    def __parse_variant(self, variant: str) -> str:
        supported_variants = {"adam", "amsgrad", "padam"}
        if (v := variant.lower().strip()) not in supported_variants:
            raise ValueError(f"Unsupported variant '{variant}'.")
        return v



# Custom warnings used inside Pyxu.


class PyxuWarning(UserWarning):
    """
    Parent class of all warnings raised in Pyxu.
    """

    pass


class AutoInferenceWarning(PyxuWarning):
    """
    Use when a quantity was auto-inferenced with possible caveats.
    """

    pass


class PerformanceWarning(PyxuWarning):
    """
    Use for performance-related warnings.
    """


class PrecisionWarning(PyxuWarning):
    """
    Use for precision-related warnings.
    """


class DenseWarning(PyxuWarning):
    """
    Use for sparse-based algos which revert to dense arrays.
    """


class NonTransparentWarning(PyxuWarning):
    """
    Inform test suite runner of (safe) non-transparent function call.
    """


class BackendWarning(PyxuWarning):
    """
    Inform user of a backend-specific problem to be aware of.
    """


class ContributionWarning(PyxuWarning):
    """
    Use for warnings related to Pyxu plugins.
    """



# How ScaleRule tests work:
#
# * ScaleRuleMixin auto-defines all arithmetic method (input,output) pairs.
#   [Caveat: we assume all tested examples are defined on \bR.] (This is not a problem in practice.)
#   [Caveat: we assume the base operators (op_orig) are correctly implemented.] (True if choosing test operators from examples/.)
#
# * To test a scaled-operator, inherit from ScaleRuleMixin and the suitable conftest.MapT subclass
#   which the scaled operator should abide by.


import collections.abc as cabc
import itertools

import numpy as np
import pytest

import pyxu.abc as pxa
import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.runtime as pxrt
import pyxu_tests.operator.conftest as conftest

op_scale_params = frozenset({-2.3, -1, 1, 2})
op_scale_unit = frozenset(_ for _ in op_scale_params if np.isclose(np.abs(_), 1))
op_scale_nonunit = frozenset(_ for _ in op_scale_params if ~np.isclose(np.abs(_), 1))
op_scale_positive = frozenset(_ for _ in op_scale_params if _ > 0)
op_scale_negative = frozenset(_ for _ in op_scale_params if _ < 0)
op_scale_nonidentity = frozenset(_ for _ in op_scale_params if ~np.isclose(_, 1))


class ScaleRuleMixin:
    # Fixtures ----------------------------------------------------------------
    @pytest.fixture
    def op_orig(self) -> pxt.OpT:
        # Override in inherited class with the operator to be scaled.
        raise NotImplementedError

    @pytest.fixture(params=op_scale_params)
    def op_scale(self, request) -> pxt.Real:
        # Scaling factors applied to op_orig()
        return request.param

    @pytest.fixture(
        params=itertools.product(
            [
                "scale_left",
                "scale_right",
                "divide_right",
            ],
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def spec(self, op_orig, op_scale, request) -> tuple[pxt.OpT, pxd.NDArrayInfo, pxrt.Width]:
        mode, ndi, width = request.param
        if mode == "scale_left":
            op = op_scale * op_orig
        elif mode == "scale_right":
            op = op_orig * op_scale
        else:  # divide_right
            op = op_orig / (1 / op_scale)
        return op, ndi, width

    @pytest.fixture
    def data_shape(self, op_orig) -> pxt.OpShape:
        return op_orig.shape

    @pytest.fixture
    def data_apply(self, op_orig, op_scale) -> conftest.DataLike:
        dim = self._sanitize(op_orig.dim, 7)
        arr = self._random_array((dim,), seed=20)  # random seed for reproducibility
        out = op_orig.apply(arr) * op_scale
        return dict(
            in_=dict(arr=arr),
            out=out,
        )

    @pytest.fixture
    def data_adjoint(self, op_orig, op_scale) -> conftest.DataLike:
        codim = self._sanitize(op_orig.codim, 7)
        arr = self._random_array((codim,), seed=20)  # random seed for reproducibility
        out = op_orig.adjoint(arr) * op_scale
        return dict(
            in_=dict(arr=arr),
            out=out,
        )

    @pytest.fixture
    def data_grad(self, op_orig, op_scale) -> conftest.DataLike:
        dim = self._sanitize(op_orig.dim, 7)
        arr = self._random_array((dim,), seed=20)  # random seed for reproducibility
        out = op_orig.grad(arr) * op_scale
        return dict(
            in_=dict(arr=arr),
            out=out,
        )

    @pytest.fixture
    def data_prox(self, op_orig, op_scale) -> conftest.DataLike:
        dim = self._sanitize(op_orig.dim, 7)
        arr = self._random_array((dim,), seed=20)  # random seed for reproducibility
        tau = np.abs(self._random_array((1,), seed=21))[0]  # random seed for reproducibility
        out = op_orig.prox(arr, tau * op_scale)
        return dict(
            in_=dict(
                arr=arr,
                tau=tau,
            ),
            out=out,
        )

    @pytest.fixture
    def data_math_lipschitz(self, op) -> cabc.Collection[np.ndarray]:
        N_test, dim = 5, self._sanitize(op.dim, 7)
        return self._random_array((N_test, dim))

    @pytest.fixture
    def data_math_diff_lipschitz(self, op) -> cabc.Collection[np.ndarray]:
        N_test, dim = 5, self._sanitize(op.dim, 7)
        return self._random_array((N_test, dim))

    # Tests -------------------------------------------------------------------
    def test_interface_asloss(self, op, xp, width, op_orig):
        self._skip_if_disabled()
        if not op_orig.has(pxa.Property.FUNCTIONAL):
            pytest.skip("asloss() unavailable for non-functionals.")

        try:
            op_orig.asloss()  # detect if fails
            super().test_interface_asloss(op, xp, width)
        except NotImplementedError:
            pytest.skip("asloss() unsupported by base operator.")


# Test classes (Maps) ---------------------------------------------------------
class TestScaleRuleMap(ScaleRuleMixin, conftest.MapT):
    @pytest.fixture
    def op_orig(self):
        import pyxu_tests.operator.examples.test_map as tc

        return tc.ReLU(M=6)


class TestScaleRuleDiffMap(ScaleRuleMixin, conftest.DiffMapT):
    @pytest.fixture
    def op_orig(self):
        import pyxu_tests.operator.examples.test_diffmap as tc

        return tc.Sin(M=6)


class TestScaleRuleLinOp(ScaleRuleMixin, conftest.LinOpT):
    @pytest.fixture
    def op_orig(self):
        import pyxu_tests.operator.examples.test_linop as tc

        return tc.Tile(N=3, M=4)


class TestScaleRuleSquareOp(ScaleRuleMixin, conftest.SquareOpT):
    @pytest.fixture
    def op_orig(self):
        import pyxu_tests.operator.examples.test_squareop as tc

        return tc.CumSum(N=7)


class TestScaleRuleNormalOp(ScaleRuleMixin, conftest.NormalOpT):
    @pytest.fixture
    def op_orig(self):
        import pyxu_tests.operator.examples.test_normalop as tc

        h = self._random_array((5,), seed=2)
        return tc.CircularConvolution(h=h)


# START UnitOp ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
class ScaleRuleUnitOp(ScaleRuleMixin):
    @pytest.fixture
    def op_orig(self):
        import pyxu_tests.operator.examples.test_unitop as tc

        return tc.Permutation(N=7)


@pytest.mark.parametrize("op_scale", op_scale_unit)
class TestScaleRuleUnitOp_UnitScale(ScaleRuleUnitOp, conftest.UnitOpT):
    pass


@pytest.mark.parametrize("op_scale", op_scale_nonunit)
class TestScaleRuleUnitOp_NonUnitScale(ScaleRuleUnitOp, conftest.NormalOpT):
    pass


# END UnitOp ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^


class TestScaleRuleSelfAdjointOp(ScaleRuleMixin, conftest.SelfAdjointOpT):
    @pytest.fixture
    def op_orig(self):
        import pyxu_tests.operator.examples.test_selfadjointop as tc

        return tc.SelfAdjointConvolution(N=7)


# START PosDefOp ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
class ScaleRulePosDefOp(ScaleRuleMixin):
    @pytest.fixture
    def op_orig(self):
        import pyxu_tests.operator.examples.test_posdefop as tc

        return tc.PSDConvolution(N=7)


@pytest.mark.parametrize("op_scale", op_scale_positive)
class TestScaleRulePosDefOp_PositiveScale(ScaleRulePosDefOp, conftest.PosDefOpT):
    pass


@pytest.mark.parametrize("op_scale", op_scale_negative)
class TestScaleRulePosDefOp_NegativeScale(ScaleRulePosDefOp, conftest.SelfAdjointOpT):
    pass


# END PosDefOp ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^


# START ProjOp ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
class ScaleRuleProjOp(ScaleRuleMixin):
    @pytest.fixture
    def op_orig(self):
        import pyxu_tests.operator.examples.test_projop as tc

        return tc.Oblique(N=7, alpha=0.3)


@pytest.mark.parametrize("op_scale", [1])
class TestScaleRuleProjOp_IdentityScale(ScaleRuleProjOp, conftest.ProjOpT):
    pass


@pytest.mark.parametrize("op_scale", op_scale_nonidentity)
class TestScaleRuleProjOp_NonIdentityScale(ScaleRuleProjOp, conftest.SquareOpT):
    pass


# END ProjOp ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^


# START OrthProjOp ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
class ScaleRuleOrthProjOp(ScaleRuleMixin):
    @pytest.fixture
    def op_orig(self):
        import pyxu_tests.operator.examples.test_orthprojop as tc

        return tc.ScaleDown(N=7)


@pytest.mark.parametrize("op_scale", [1])
class TestScaleRuleOrthProjOp_IdentityScale(ScaleRuleOrthProjOp, conftest.OrthProjOpT):
    pass


@pytest.mark.parametrize("op_scale", op_scale_nonidentity)
class TestScaleRuleOrthProjOp_NonIdentityScale(ScaleRuleOrthProjOp, conftest.SelfAdjointOpT):
    pass


# END OrthProjOp ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^


# Test classes (Funcs) --------------------------------------------------------
class TestScaleRuleFunc(ScaleRuleMixin, conftest.FuncT):
    @pytest.fixture(params=[7])
    def op_orig(self, request):
        import pyxu_tests.operator.examples.test_func as tc

        return tc.Median(dim=request.param)


class TestScaleRuleDiffFunc(ScaleRuleMixin, conftest.DiffFuncT):
    @pytest.fixture(params=[7])
    def op_orig(self, request):
        import pyxu_tests.operator.examples.test_difffunc as tc

        return tc.SquaredL2Norm(M=request.param)


# START ProxFunc ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
class ScaleRuleProxFunc(ScaleRuleMixin):
    @pytest.fixture(params=[7])
    def op_orig(self, request):
        import pyxu_tests.operator.examples.test_proxfunc as tc

        return tc.L1Norm(M=request.param)


@pytest.mark.parametrize("op_scale", op_scale_positive)
class TestScaleRuleProxFunc_PositiveScale(ScaleRuleProxFunc, conftest.ProxFuncT):
    pass


@pytest.mark.parametrize("op_scale", op_scale_negative)
class TestScaleRuleProxFunc_NegativeScale(ScaleRuleProxFunc, conftest.FuncT):
    pass


# END ProxFunc ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^


# START ProxDiffFunc ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
class ScaleRuleProxDiffFunc(ScaleRuleMixin):
    @pytest.fixture(params=[7])
    def op_orig(self, request):
        import pyxu_tests.operator.examples.test_proxdifffunc as tc

        return tc.SquaredL2Norm(M=request.param)


@pytest.mark.parametrize("op_scale", op_scale_positive)
class TestScaleRuleProxDiffFunc_PositiveScale(ScaleRuleProxDiffFunc, conftest.ProxDiffFuncT):
    pass


@pytest.mark.parametrize("op_scale", op_scale_negative)
class TestScaleRuleProxDiffFunc_NegativeScale(ScaleRuleProxDiffFunc, conftest.DiffFuncT):
    pass


# END ProxDiffFunc ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^


# START QuadraticFunc ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
class ScaleRuleQuadraticFunc(ScaleRuleMixin):
    @pytest.fixture(params=[0, 1])
    def op_orig(self, request):
        from pyxu_tests.operator.examples.test_linfunc import ScaledSum
        from pyxu_tests.operator.examples.test_posdefop import PSDConvolution

        N = 7
        op = {
            0: pxa.QuadraticFunc(shape=(1, N)),
            1: pxa.QuadraticFunc(
                shape=(1, N),
                Q=PSDConvolution(N=N),
                c=ScaledSum(N=N),
                t=1,
            ),
        }[request.param]
        return op


@pytest.mark.parametrize("op_scale", op_scale_positive)
class TestScaleRuleQuadraticFunc_PositiveScale(ScaleRuleQuadraticFunc, conftest.QuadraticFuncT):
    pass


@pytest.mark.parametrize("op_scale", op_scale_negative)
class TestScaleRuleQuadraticFunc_NegativeScale(ScaleRuleQuadraticFunc, conftest.DiffFuncT):
    pass


# END QuadraticFunc ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^


class TestScaleRuleLinFunc(ScaleRuleMixin, conftest.LinFuncT):
    @pytest.fixture()
    def op_orig(self):
        import pyxu_tests.operator.examples.test_linfunc as tc

        return tc.ScaledSum(N=7)



import numpy as np

import pyxu.abc as pxa
import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.operator.interop.source as px_src
import pyxu.runtime as pxrt
import pyxu.util as pxu

__all__ = [
    "kron",
    "khatri_rao",
]


def kron(A: pxt.OpT, B: pxt.OpT) -> pxt.OpT:
    r"""
    `Kronecker product <https://en.wikipedia.org/wiki/Kronecker_product#Definition>`_ :math:`A \otimes B` between two
    linear operators.

    The Kronecker product :math:`A \otimes B` is defined as

    .. math::

       A \otimes B
       =
       \left[
           \begin{array}{ccc}
               A_{11} B     & \cdots & A_{1N_{A}} B \\
               \vdots       & \ddots & \vdots   \\
               A_{M_{A}1} B & \cdots & A_{M_{A}N_{A}} B \\
           \end{array}
       \right],

    where :math:`A : \mathbb{R}^{N_{A}} \to \mathbb{R}^{M_{A}}`, and :math:`B : \mathbb{R}^{N_{B}} \to
    \mathbb{R}^{M_{B}}`.

    Parameters
    ----------
    A: OpT
        (mA, nA) linear operator
    B: OpT
        (mB, nB) linear operator

    Returns
    -------
    op: OpT
        (mA*mB, nA*nB) linear operator.

    Notes
    -----
    This implementation is **matrix-free** by leveraging properties of the Kronecker product, i.e.  :math:`A` and
    :math:`B` need not be known explicitly.  In particular :math:`(A \otimes B) x` and :math:`(A \otimes B)^{*} x` are
    computed implicitly via the relation:

    .. math::

       \text{vec}(\mathbf{A}\mathbf{B}\mathbf{C})
       =
       (\mathbf{C}^{T} \otimes \mathbf{A}) \text{vec}(\mathbf{B}),

    where :math:`\mathbf{A}`, :math:`\mathbf{B}`, and :math:`\mathbf{C}` are matrices.
    """

    def _infer_op_shape(shA: pxt.OpShape, shB: pxt.OpShape) -> pxt.OpShape:
        sh = (shA[0] * shB[0], shA[1] * shB[1])
        return sh

    def _infer_op_klass(A: pxt.OpT, B: pxt.OpT) -> pxt.OpC:
        # linear \kron linear -> linear
        # square (if output square)
        # normal \kron normal -> normal
        # unit \kron unit -> unit
        # self-adj \kron self-adj -> self-adj
        # pos-def \kron pos-def -> pos-def
        # idemp \kron idemp -> idemp
        # func \kron func -> func
        properties = set(A.properties() & B.properties())
        sh = _infer_op_shape(A.shape, B.shape)
        if sh[0] == sh[1]:
            properties.add(pxa.Property.LINEAR_SQUARE)
        if pxa.Property.FUNCTIONAL in properties:
            klass = pxa.LinFunc
        else:
            klass = pxa.Operator._infer_operator_type(properties)
        return klass

    @pxrt.enforce_precision(i="arr")
    def op_apply(_, arr: pxt.NDArray) -> pxt.NDArray:
        # If `x` is a vector, then:
        #     (A \kron B)(x) = vec(B * mat(x) * A.T)
        sh_prefix = arr.shape[:-1]
        sh_dim = len(sh_prefix)

        x = arr.reshape((*sh_prefix, _._A.dim, _._B.dim))  # (..., A.dim, B.dim)
        y = _._B.apply(x)  # (..., A.dim, B.codim)
        z = y.transpose((*range(sh_dim), -1, -2))  # (..., B.codim, A.dim)
        t = _._A.apply(z)  # (..., B.codim, A.codim)
        u = t.transpose((*range(sh_dim), -1, -2))  # (..., A.codim, B.codim)

        out = u.reshape((*sh_prefix, -1))  # (..., A.codim * B.codim)
        return out

    @pxrt.enforce_precision(i="arr")
    def op_adjoint(_, arr: pxt.NDArray) -> pxt.NDArray:
        # If `x` is a vector, then:
        #     (A \kron B).H(x) = vec(B.H * mat(x) * A.conj)
        sh_prefix = arr.shape[:-1]
        sh_dim = len(sh_prefix)

        x = arr.reshape((*sh_prefix, _._A.codim, _._B.codim))  # (..., A.codim, B.codim)
        y = _._B.adjoint(x)  # (..., A.codim, B.dim)
        z = y.transpose((*range(sh_dim), -1, -2))  # (..., B.dim, A.codim)
        t = _._A.adjoint(z)  # (..., B.dim, A.dim)
        u = t.transpose((*range(sh_dim), -1, -2))  # (..., A.dim, B.dim)

        out = u.reshape((*sh_prefix, -1))  # (..., A.dim * B.dim)
        return out

    def op_estimate_lipschitz(_, **kwargs) -> pxt.Real:
        no_eval = "__rule" in kwargs
        if no_eval:
            L_A = _._A.lipschitz
            L_B = _._B.lipschitz
            L = L_A * L_B
        else:
            L = _.__class__.estimate_lipschitz(_, **kwargs)
        return L

    def op_asarray(_, **kwargs) -> pxt.NDArray:
        # (A \kron B).asarray() = A.asarray() \kron B.asarray()
        A = _._A.asarray(**kwargs)
        B = _._B.asarray(**kwargs)
        xp = kwargs.get("xp", pxd.NDArrayInfo.default().module())
        C = xp.tensordot(A, B, axes=0).transpose((0, 2, 1, 3)).reshape(_.shape)
        return C

    def op_gram(_) -> pxt.OpT:
        # (A \kron B).gram() = A.gram() \kron B.gram()
        A = _._A.gram()
        B = _._B.gram()
        op = kron(A, B)
        return op

    def op_cogram(_) -> pxt.OpT:
        # (A \kron B).cogram() = A.cogram() \kron B.cogram()
        A = _._A.cogram()
        B = _._B.cogram()
        op = kron(A, B)
        return op

    def op_svdvals(_, **kwargs) -> pxt.NDArray:
        # (A \kron B).svdvals(k, which)
        # = outer(
        #     A.svdvals(k, which),
        #     B.svdvals(k, which)
        #   ).[top|bottom](k)
        k = kwargs.get("k", 1)
        which = kwargs.get("which", "LM")
        if which.upper() == "SM":
            # `scipy.sparse.linalg.svds()` and `scipy.linalg.svd()` will only return up to
            # min(shape) singular values.
            # As such (A.svdvals(), B.svdvals()) output is insufficient to infer SM-singular values
            # of (A \kron B).
            D_C = _.__class__.svdvals(_, **kwargs)
        else:
            D_A = _._A.svdvals(**kwargs)
            D_B = _._B.svdvals(**kwargs)
            xp = pxu.get_array_module(D_A)
            pad_length = np.fmax(k - len(D_A) * len(D_B), 0)
            D_C = xp.concatenate(
                [
                    xp.zeros(pad_length, dtype=D_A.dtype),
                    xp.sort(xp.outer(D_A, D_B), axis=None),
                ]
            )[-k:]
        return D_C

    @pxrt.enforce_precision(i=("arr", "damp"))
    def op_pinv(_, arr: pxt.NDArray, damp: pxt.Real, **kwargs) -> pxt.NDArray:
        if np.isclose(damp, 0):
            # (A \kron B).dagger() = A.dagger() \kron B.dagger()
            op_d = kron(_._A.dagger(damp, **kwargs), _._B.dagger(damp, **kwargs))
            out = op_d.apply(arr)
        else:
            # default algorithm
            out = _.__class__.pinv(_, arr, damp, **kwargs)
        return out

    @pxrt.enforce_precision()
    def op_trace(_, **kwargs) -> pxt.Real:
        # tr(A \kron B) = tr(A) * tr(B)
        # [if both square, else default algorithm]
        P = pxa.Property.LINEAR_SQUARE
        if not _.has(P):
            raise NotImplementedError

        if _._A.has(P) and _._B.has(P):
            tr = _._A.trace(**kwargs) * _._B.trace(**kwargs)
        else:
            tr = _.__class__.trace(_, **kwargs)
        return tr

    _A = A.squeeze()
    _B = B.squeeze()
    assert (klass := _infer_op_klass(_A, _B)).has(pxa.Property.LINEAR)
    is_scalar = lambda _: _.shape == (1, 1)
    if is_scalar(_A) and is_scalar(_B):
        from pyxu.operator.linop.base import HomothetyOp

        return HomothetyOp(cst=(_A.asarray() * _B.asarray()).item(), dim=1)
    elif is_scalar(_A) and (not is_scalar(_B)):
        return _A.asarray().item() * _B
    elif (not is_scalar(_A)) and is_scalar(B):
        return _A * _B.asarray().item()
    else:
        op = px_src.from_source(
            cls=klass,
            shape=_infer_op_shape(_A.shape, _B.shape),
            embed=dict(
                _name="kron",
                _A=_A,
                _B=_B,
            ),
            apply=op_apply,
            adjoint=op_adjoint,
            asarray=op_asarray,
            gram=op_gram,
            cogram=op_cogram,
            svdvals=op_svdvals,
            pinv=op_pinv,
            trace=op_trace,
            estimate_lipschitz=op_estimate_lipschitz,
            _expr=lambda _: (_._name, _._A, _._B),
        )
        op.lipschitz = op.estimate_lipschitz(__rule=True)
    return op


def khatri_rao(A: pxt.OpT, B: pxt.OpT) -> pxt.OpT:
    r"""
    `Column-wise Khatri-Rao product
    <https://en.wikipedia.org/wiki/Khatri%E2%80%93Rao_product#Column-wise_Kronecker_product>`_ :math:`A \circ B` between
    two linear operators.

    The Khatri-Rao product :math:`A \circ B` is defined as

    .. math::

       A \circ B
       =
       \left[
           \begin{array}{ccc}
           \mathbf{a}_{1} \otimes \mathbf{b}_{1} & \cdots & \mathbf{a}_{N} \otimes \mathbf{b}_{N}
           \end{array}
       \right],

    where :math:`A : \mathbb{R}^{N} \to \mathbb{R}^{M_{A}}`, :math:`B : \mathbb{R}^{N} \to \mathbb{R}^{M_{B}}`, and
    :math:`\mathbf{a}_{k}` (repectively :math:`\mathbf{b}_{k}`) denotes the :math:`k`-th column of :math:`A`
    (respectively :math:`B`).

    Parameters
    ----------
    A: OpT
        (mA, n) linear operator
    B: OpT
        (mB, n) linear operator

    Returns
    -------
    op: OpT
        (mA*mB, n) linear operator.

    Notes
    -----
    This implementation is **matrix-free** by leveraging properties of the Khatri-Rao product, i.e.  :math:`A` and
    :math:`B` need not be known explicitly.  In particular :math:`(A \circ B) x` and :math:`(A \circ B)^{*} x` are
    computed implicitly via the relation:

    .. math::

       \text{vec}(\mathbf{A}\text{diag}(\mathbf{b})\mathbf{C})
       =
       (\mathbf{C}^{T} \circ \mathbf{A}) \mathbf{b},

    where :math:`\mathbf{A}`, :math:`\mathbf{C}` are matrices, and :math:`\mathbf{b}` is a vector.

    Note however that a matrix-free implementation of the Khatri-Rao product does not permit the same optimizations as a
    matrix-based implementation.  Thus the Khatri-Rao product as implemented here is only marginally more efficient than
    applying :py:func:`~pyxu.operator.kron` and pruning its output.
    """

    def _infer_op_shape(shA: pxt.OpShape, shB: pxt.OpShape) -> pxt.OpShape:
        if shA[1] != shB[1]:
            raise ValueError(f"Khatri-Rao product of {shA} and {shB} operators forbidden.")
        sh = (shA[0] * shB[0], shA[1])
        return sh

    def _infer_op_klass(A: pxt.OpT, B: pxt.OpT) -> pxt.OpC:
        # linear \kr linear -> linear
        # square (if output square)
        sh = _infer_op_shape(A.shape, B.shape)
        if sh[0] == 1:
            klass = pxa.LinFunc
        else:
            properties = set(pxa.LinOp.properties())
            if sh[0] == sh[1]:
                properties.add(pxa.Property.LINEAR_SQUARE)
            klass = pxa.Operator._infer_operator_type(properties)
        return klass

    @pxrt.enforce_precision(i="arr")
    def op_apply(_, arr: pxt.NDArray) -> pxt.NDArray:
        # If `x` is a vector, then:
        #     (A \kr B)(x) = vec(B * diag(x) * A.T)
        sh_prefix = arr.shape[:-1]
        sh_dim = len(sh_prefix)
        xp = pxu.get_array_module(arr)
        I = xp.eye(N=_.dim, dtype=arr.dtype)  # noqa: E741

        x = arr.reshape((*sh_prefix, 1, _.dim))  # (..., 1, dim)
        y = _._B.apply(x * I)  # (..., dim, B.codim)
        z = y.transpose((*range(sh_dim), -1, -2))  # (..., B.codim, dim)
        t = _._A.apply(z)  # (..., B.codim, A.codim)
        u = t.transpose((*range(sh_dim), -1, -2))  # (..., A.codim, B.codim)

        out = u.reshape((*sh_prefix, -1))  # (..., A.codim * B.codim)
        return out

    @pxrt.enforce_precision(i="arr")
    def op_adjoint(_, arr: pxt.NDArray) -> pxt.NDArray:
        # If `x` is a vector, then:
        #     (A \kr B).H(x) = diag(B.H * mat(x) * A.conj)
        sh_prefix = arr.shape[:-1]
        sh_dim = len(sh_prefix)
        xp = pxu.get_array_module(arr)
        I = xp.eye(N=_.dim, dtype=arr.dtype)  # noqa: E741

        x = arr.reshape((*sh_prefix, _._A.codim, _._B.codim))  # (..., A.codim, B.codim)
        y = _._B.adjoint(x)  # (..., A.codim, B.dim)
        z = y.transpose((*range(sh_dim), -1, -2))  # (..., dim, A.codim)
        t = pxu.copy_if_unsafe(_._A.adjoint(z))  # (..., dim, dim)
        t *= I

        out = t.sum(axis=-1)  # (..., dim)
        return out

    def op_asarray(_, **kwargs) -> pxt.NDArray:
        # (A \kr B).asarray()[:,i] = A.asarray()[:,i] \kron B.asarray()[:,i]
        A = _._A.asarray(**kwargs).T.reshape((_.dim, _._A.codim, 1))
        B = _._B.asarray(**kwargs).T.reshape((_.dim, 1, _._B.codim))
        C = (A * B).reshape((_.dim, -1)).T
        return C

    @pxrt.enforce_precision()
    def op_lipschitz(_, **kwargs) -> pxt.Real:
        if kwargs.get("tight", False):
            _._lipschitz = _.__class__.lipschitz(_, **kwargs)
        else:
            op = kron(_._A, _._B)
            _._lipschitz = op.lipschitz(**kwargs)
        return _._lipschitz

    _A = A.squeeze()
    _B = B.squeeze()
    assert (klass := _infer_op_klass(_A, _B)).has(pxa.Property.LINEAR)

    op = px_src.from_source(
        cls=klass,
        shape=_infer_op_shape(_A.shape, _B.shape),
        embed=dict(
            _name="khatri_rao",
            _A=_A,
            _B=_B,
        ),
        apply=op_apply,
        adjoint=op_adjoint,
        asarray=op_asarray,
        _expr=lambda _: (_._name, _._A, _._B),
    )

    # kr(A,B) = kron(A,B) + sub-sampling -> upper-bound provided by kron(A,B).lipschitz
    op.lipschitz = kron(_A, _B).lipschitz
    return op



# How hstack() tests work:
#
# * HStackMixin auto-defines all (input,output) pairs.
#   [Caveat: we assume the base operators (op_all) are correctly implemented.]
#   (True if choosing test operators from examples/.)
#
# * To test a hstack-ed operator, inherit from HStackMixin and the suitable conftest.MapT subclass
#   which the compound operator should abide by.

import collections.abc as cabc
import itertools

import numpy as np
import pytest

import pyxu.abc as pxa
import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.operator as pxo
import pyxu.runtime as pxrt
import pyxu_tests.operator.conftest as conftest


def spec_op_map(klass: pxt.OpC, N: int = 2) -> list[list[pxt.OpT]]:
    # create all possible (op1, ..., opN) tuples which, when used to create a hstack-ed operator,
    # produce an operator of type `klass`. (Necessarily a non-functional.)
    def op_map(dim: int):
        import pyxu_tests.operator.examples.test_map as tc

        return tc.ReLU(M=dim)

    def op_diffmap(dim: int):
        import pyxu_tests.operator.examples.test_diffmap as tc

        return tc.Sin(M=dim)

    def op_linop(dim: int, codim_scale: int):
        import pyxu_tests.operator.examples.test_linop as tc

        return tc.Tile(N=dim, M=codim_scale)

    def op_squareop(dim: int):
        import pyxu_tests.operator.examples.test_squareop as tc

        return tc.CumSum(N=dim)

    def op_normalop(dim: int):
        import pyxu_tests.operator.examples.test_normalop as tc

        rng = np.random.default_rng(seed=2)
        h = rng.normal(size=(dim,))
        return tc.CircularConvolution(h=h)

    def op_unitop(dim: int):
        import pyxu_tests.operator.examples.test_unitop as tc

        return tc.Permutation(N=dim)

    def op_selfadjointop(dim: int):
        import pyxu_tests.operator.examples.test_selfadjointop as tc

        return tc.SelfAdjointConvolution(N=dim)

    def op_posdefop(dim: int):
        import pyxu_tests.operator.examples.test_posdefop as tc

        return tc.PSDConvolution(N=dim)

    def op_projop(dim: int):
        import pyxu_tests.operator.examples.test_projop as tc

        return tc.Oblique(N=dim, alpha=np.pi / 4)

    def op_orthprojop(dim: int):
        import pyxu_tests.operator.examples.test_orthprojop as tc

        return tc.ScaleDown(N=dim)

    def condition(ops: list[pxt.OpT], klass: pxt.OpC) -> bool:
        # Return true if hstack(ops) forms a klass object. [Not a sub-type]
        op_sum = ops[0]  # hack to get the output type (modulo square-ness)
        for op in ops[1:]:
            op_sum = op_sum + op
        properties = op_sum.properties()

        codim = ops[0].codim
        dim = sum(op.dim for op in ops)
        if (pxa.Property.LINEAR in properties) and (codim == dim):
            properties = pxa.SquareOp.properties()
        else:
            properties = properties & pxa.LinOp.properties()

        _klass = pxa.Operator._infer_operator_type(properties)
        return _klass == klass

    ops = []
    for _ops in itertools.combinations_with_replacement(
        [
            op_map(7),
            op_diffmap(7),
            op_linop(7, 1),
            op_squareop(7),
            op_normalop(7),
            op_unitop(7),
            op_selfadjointop(7),
            op_posdefop(7),
            op_projop(7),
            op_orthprojop(7),
        ],
        N,
    ):
        if condition(_ops, klass):
            ops.append(_ops)
    return ops


def spec_op_func(klass: pxt.OpC, N: int = 2) -> list[list[pxt.OpT]]:
    # create all possible (op1, ..., opN) tuples which, when used to create a hstack-ed operator,
    # produce an operator of type `klass`. (Necessarily a functional.)
    def op_difffunc(dim: int):
        import pyxu_tests.operator.examples.test_difffunc as tc

        return tc.SquaredL2Norm(M=dim)

    def op_proxfunc(dim: int):
        import pyxu_tests.operator.examples.test_proxfunc as tc

        return tc.L1Norm(M=dim)

    def op_proxdifffunc(dim: int):
        import pyxu_tests.operator.examples.test_proxdifffunc as tc

        return tc.SquaredL2Norm(M=dim)

    def op_quadraticfunc(dim: int):
        from pyxu_tests.operator.examples.test_linfunc import ScaledSum
        from pyxu_tests.operator.examples.test_posdefop import PSDConvolution

        return pxa.QuadraticFunc(
            shape=(1, dim),
            Q=PSDConvolution(N=dim),
            c=ScaledSum(N=dim),
            t=1,
        )

    def op_linfunc(dim: int):
        import pyxu_tests.operator.examples.test_linfunc as tc

        op = tc.ScaledSum(N=dim)
        return op

    def condition(ops: list[pxt.OpT], klass: pxt.OpC) -> bool:
        # Return true if hstack(ops) forms a klass object. [Not a sub-type]
        op_sum = ops[0]  # hack to get the output type (modulo square-ness)
        for op in ops[1:]:
            op_sum = op_sum + op
        _klass = pxa.Operator._infer_operator_type(op_sum.properties())
        return _klass == klass

    ops = []
    for _ops in itertools.combinations_with_replacement(
        [
            op_difffunc(9),
            op_proxfunc(9),
            op_proxdifffunc(9),
            op_quadraticfunc(9),
            op_linfunc(9),
        ],
        N,
    ):
        if condition(_ops, klass):
            ops.append(_ops)
    return ops


# Data Mixin ------------------------------------------------------------------
class HStackMixin:
    # Fixtures ----------------------------------------------------------------
    @pytest.fixture
    def op_all(self) -> list[pxt.OpT]:
        # Override in inherited class with sub-operators.
        raise NotImplementedError

    @pytest.fixture(
        params=itertools.product(
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def spec(self, op_all, request) -> tuple[pxt.OpT, pxd.NDArrayInfo, pxrt.Width]:
        ndi, width = request.param
        op = pxo.hstack(op_all)
        return op, ndi, width

    @pytest.fixture
    def data_shape(self, op_all) -> pxt.OpShape:
        dim = sum(op.dim for op in op_all)
        codim = op_all[0].codim
        sh = (codim, dim)
        return sh

    @pytest.fixture
    def data_apply(self, op_all) -> conftest.DataLike:
        parts_arr = []
        parts_out = []
        for op in op_all:
            p_arr = self._random_array((op.dim,), seed=3)  # random seed for reproducibility
            parts_arr.append(p_arr)
            p_out = op.apply(p_arr)
            parts_out.append(p_out)
        arr = np.concatenate(parts_arr)
        out = np.sum(parts_out, axis=0)
        return dict(
            in_=dict(arr=arr),
            out=out,
        )

    @pytest.fixture
    def data_adjoint(self, op_all) -> conftest.DataLike:
        arr = self._random_array((op_all[0].codim,), seed=4)  # random seed for reproducibility
        parts_out = []
        for op in op_all:
            p_out = op.adjoint(arr)
            parts_out.append(p_out)
        out = np.concatenate(parts_out)
        return dict(
            in_=dict(arr=arr),
            out=out,
        )

    @pytest.fixture
    def data_grad(self, op_all) -> conftest.DataLike:
        parts_arr = []
        parts_out = []
        for op in op_all:
            p_arr = self._random_array((op.dim,), seed=5)  # random seed for reproducibility
            parts_arr.append(p_arr)
            p_out = op.grad(p_arr)
            parts_out.append(p_out)
        arr = np.concatenate(parts_arr)
        out = np.concatenate(parts_out)
        return dict(
            in_=dict(arr=arr),
            out=out,
        )

    @pytest.fixture
    def data_prox(self, op_all) -> conftest.DataLike:
        parts_arr = []
        parts_out = []
        tau = np.abs(self._random_array((1,), seed=30))[0]  # random seed for reproducibility
        for op in op_all:
            p_arr = self._random_array((op.dim,), seed=6)  # random seed for reproducibility
            parts_arr.append(p_arr)
            p_out = op.prox(arr=p_arr, tau=tau)
            parts_out.append(p_out)
        arr = np.concatenate(parts_arr)
        out = np.concatenate(parts_out)
        return dict(
            in_=dict(
                arr=arr,
                tau=tau,
            ),
            out=out,
        )

    @pytest.fixture
    def data_math_lipschitz(self, data_shape) -> cabc.Collection[np.ndarray]:
        N_test, dim = 5, data_shape[1]
        return self._random_array((N_test, dim))

    @pytest.fixture
    def data_math_diff_lipschitz(self, data_shape) -> cabc.Collection[np.ndarray]:
        N_test, dim = 5, data_shape[1]
        return self._random_array((N_test, dim))


class HStackFuncMixin(HStackMixin):
    # Tests -------------------------------------------------------------------
    def test_interface_asloss(self, op, xp, width):
        self._skip_if_disabled()
        with pytest.raises(NotImplementedError):
            # asloss() is ambiguous for hstack-ed operators -> must fail.
            op.asloss()


# Test classes (Maps) ---------------------------------------------------------
class TestHStackMap(HStackMixin, conftest.MapT):
    @pytest.fixture(params=spec_op_map(pxa.Map))
    def op_all(self, request):
        return request.param


class TestHStackDiffMap(HStackMixin, conftest.DiffMapT):
    @pytest.fixture(params=spec_op_map(pxa.DiffMap))
    def op_all(self, request):
        return request.param


class TestHStackLinOp(HStackMixin, conftest.LinOpT):
    @pytest.fixture(params=spec_op_map(pxa.LinOp))
    def op_all(self, request):
        return request.param


class TestHStackSquareOp(HStackMixin, conftest.SquareOpT):
    @pytest.fixture(params=spec_op_map(pxa.SquareOp))
    def op_all(self, request):
        return request.param


# Test classes (Funcs) --------------------------------------------------------
class TestHStackFunc(HStackFuncMixin, conftest.FuncT):
    @pytest.fixture(params=spec_op_func(pxa.Func))
    def op_all(self, request):
        return request.param


class TestHStackProxFunc(HStackFuncMixin, conftest.ProxFuncT):
    @pytest.fixture(params=spec_op_func(pxa.ProxFunc))
    def op_all(self, request):
        return request.param


class TestHStackDiffFunc(HStackFuncMixin, conftest.DiffFuncT):
    @pytest.fixture(params=spec_op_func(pxa.DiffFunc))
    def op_all(self, request):
        return request.param


class TestHStackProxDiffFunc(HStackFuncMixin, conftest.ProxDiffFuncT):
    @pytest.fixture(params=spec_op_func(pxa.ProxDiffFunc))
    def op_all(self, request):
        return request.param


class TestHStackQuadraticFunc(HStackFuncMixin, conftest.QuadraticFuncT):
    @pytest.fixture(params=spec_op_func(pxa.QuadraticFunc))
    def op_all(self, request):
        return request.param


class TestHStackLinFunc(HStackFuncMixin, conftest.LinFuncT):
    @pytest.fixture(params=spec_op_func(pxa.LinFunc))
    def op_all(self, request):
        return request.param



import dask.array as da
import distributed  # noqa: F401
import numpy as np

import pyxu.operator as pxo
import pyxu.runtime as pxrt
import pyxu.util as pxu

# client = distributed.Client(processes=False)
use_dask = True


def NUFFT2_array(x, N, isign) -> np.ndarray:
    D = x.shape[-1]
    if isinstance(N, int):
        N = (N,) * D
    A = np.meshgrid(
        *[np.arange(-(n // 2), (n - 1) // 2 + 1) for n in N],
        indexing="ij",
    )
    B = np.stack(A, axis=0).reshape((D, -1)).T
    return np.exp(1j * np.sign(isign) * B @ x.T).T


rng = np.random.default_rng(0)
D, M, N = 3, 200, (5, 3, 4)
x = rng.normal(size=(M, D))

with pxrt.Precision(pxrt.Width.DOUBLE):
    N_trans, isign, real = 10, -1, True
    A = pxo.NUFFT.type2(
        x=x,
        N=N,
        n_trans=N_trans,
        isign=isign,
        eps=1e-6,
        real=real,
        debug=2,
    )
    B = NUFFT2_array(x, N, isign)

    sh_pad = (2, 3, 4)
    arr = rng.normal(size=(*sh_pad, *N))
    if not real:
        arr = arr + 1j * rng.normal(size=arr.shape)
    if use_dask:
        arr = da.array(arr)

    A_out_fw = pxu.view_as_complex(A.apply(pxu.view_as_real(arr.reshape(*sh_pad, -1))))
    B_out_fw = np.tensordot((arr.compute() if use_dask else arr).reshape(*sh_pad, -1), B, axes=[[-1], [-1]])

    A_out_bw = A.adjoint(pxu.view_as_real(A_out_fw))
    if not real:
        A_out_bw = pxu.view_as_complex(A_out_bw)
    A_out_bw = A_out_bw.reshape(*sh_pad, *N)
    B_out_bw = np.tensordot(B_out_fw, B.conj().T, axes=[[-1], [-1]]).reshape(*sh_pad, *N)
    if real:
        B_out_bw = B_out_bw.real

    res_fw = (np.linalg.norm(A_out_fw - B_out_fw, axis=-1) / np.linalg.norm(B_out_fw, axis=-1)).max()
    res_bw = (
        np.linalg.norm((A_out_bw - B_out_bw).reshape(*sh_pad, -1), axis=-1)
        / np.linalg.norm(B_out_bw.reshape(*sh_pad, -1), axis=-1)
    ).max()
    print(float(res_fw))
    print(float(res_bw))



import collections.abc as cabc
import functools
import operator
import typing as typ
import warnings

import numpy as np

import pyxu.abc as pxa
import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.info.warning as pxw
import pyxu.runtime as pxrt
import pyxu.util as pxu
from pyxu.operator.linop.pad import Pad
from pyxu.operator.linop.select import Trim
from pyxu.operator.linop.stencil._stencil import _Stencil

__all__ = [
    "Stencil",
    "Correlate",
    "Convolve",
]


class Stencil(pxa.SquareOp):
    r"""
    Multi-dimensional JIT-compiled stencil.

    Stencils are a common computational pattern in which array elements are updated according to some fixed pattern
    called the *stencil kernel*.  Notable examples include multi-dimensional convolutions, correlations and finite
    differences.  (See Notes for a definition.)

    Stencils can be evaluated efficiently on CPU/GPU architectures.

    Several boundary conditions are supported.  Moreover boundary conditions may differ per axis.

    .. rubric:: Implementation Notes

    * Numba (and its ``@stencil`` decorator) is used behind the scenes to compile efficient machine code from a stencil
      kernel specification.  This has 2 consequences:

      * :py:class:`~pyxu.operator.Stencil` instances are **not arraymodule-agnostic**: they will only work with NDArrays
         belonging to the same array module as `kernel`.
      * Compiled stencils are not **precision-agnostic**: they will only work on NDArrays with the same dtype as
        `kernel`.  A warning is emitted if inputs must be cast to the kernel dtype.

    * Stencil kernels can be specified in two forms: (See :py:meth:`~pyxu.operator.Stencil.__init__` for details.)

      * A single non-separable :math:`D`-dimensional kernel :math:`k[i_{1},\ldots,i_{D}]` of shape
        :math:`(K_{1},\ldots,K_{D})`.
      * A sequence of separable :math:`1`-dimensional kernel(s) :math:`k_{d}[i]` of shapes
        :math:`(K_{1},),\ldots,(K_{D},)` such that :math:`k[i_{1},\ldots,i_{D}] = \Pi_{d=1}^{D} k_{d}[i_{d}]`.

    .. rubric:: Mathematical Notes

    Given a :math:`D`-dimensional array :math:`x\in\mathbb{R}^{N_1\times\cdots\times N_D}` and kernel
    :math:`k\in\mathbb{R}^{K_1\times\cdots\times K_D}` with center :math:`(c_1, \ldots, c_D)`, the output of the stencil
    operator is an array :math:`y\in\mathbb{R}^{N_1\times\cdots\times N_D}` given by:

    .. math::

       y[i_{1},\ldots,i_{D}]
       =
       \sum_{q_{1},\ldots,q_{D}=0}^{K_{1},\ldots,K_{D}}
       x[i_{1} - c_{1} + q_{1},\ldots,i_{D} - c_{D} + q_{D}]
       \,\cdot\,
       k[q_{1},\ldots,q_{D}].

    This corresponds to a *correlation* with a shifted version of the kernel :math:`k`.

    Numba stencils assume summation terms involving out-of-bound indices of :math:`x` are set to zero.
    :py:class:`~pyxu.operator.Stencil` lifts this constraint by extending the stencil to boundary values via pre-padding
    and post-trimming.  Concretely, any stencil operator :math:`S` instantiated with :py:class:`~pyxu.operator.Stencil`
    can be written as the composition :math:`S = TS_0P`, where :math:`(T, S_0, P)` are trimming, stencil with
    zero-padding conditions, and padding operators respectively.  This construct allows
    :py:class:`~pyxu.operator.Stencil` to handle complex boundary conditions under which :math:`S` *may not be a proper
    stencil* (i.e., varying kernel) but can still be implemented efficiently via a proper stencil upon appropriate
    trimming/padding.

    For example consider the decomposition of the following (improper) stencil operator:

    .. code-block:: python3

       >>> S = Stencil(
       ...      arg_shape=(5,),
       ...      kernel=np.r_[1, 2, -3],
       ...      center=(2,),
       ...      mode="reflect",
       ... )

       >>> S.asarray()
       [[-3   2   1   0   0]
        [ 2  -2   0   0   0]
        [ 1   2  -3   0   0]
        [ 0   1   2  -3   0]
        [ 0   0   1   2  -3]] # Improper stencil (kernel varies across rows)
       =
       [[0  0  1  0  0  0  0  0  0]
        [0  0  0  1  0  0  0  0  0]
        [0  0  0  0  1  0  0  0  0]
        [0  0  0  0  0  1  0  0  0]
        [0  0  0  0  0  0  1  0  0]]  # Trimming
       @
       [[-3   0   0   0   0   0   0   0   0]
        [ 2  -3   0   0   0   0   0   0   0]
        [ 1   2  -3   0   0   0   0   0   0]
        [ 0   1   2  -3   0   0   0   0   0]
        [ 0   0   1   2  -3   0   0   0   0]
        [ 0   0   0   1   2  -3   0   0   0]
        [ 0   0   0   0   1   2  -3   0   0]
        [ 0   0   0   0   0   1   2  -3   0]
        [ 0   0   0   0   0   0   1   2  -3]]  # Proper stencil (Toeplitz structure)
       @
       [[0  0  1  0  0]
        [0  1  0  0  0]
        [1  0  0  0  0]
        [0  1  0  0  0]
        [0  0  1  0  0]
        [0  0  0  1  0]
        [0  0  0  0  1]
        [0  0  0  1  0]
        [0  0  1  0  0]]  # Padding with reflect mode.

    Note that the adjoint of a stencil operator may not necessarily be a stencil operator, or the associated center and
    boundary conditions may be hard to predict.  For example, the adjoint of the stencil operator defined above is given
    by:

    .. code-block:: python3

       >>> S.T.asarray()
       [[-3   2   1   0   0],
        [ 2  -2   2   1   0],
        [ 1   0  -3   2   1],
        [ 0   0   0  -3   2],
        [ 0   0   0   0  -3]]

    which resembles a stencil with time-reversed kernel, but with weird (if not improper) boundary conditions.  This can
    also be seen from the fact that :math:`S^\ast = P^\ast S_0^\ast T^\ast = P^\ast S_0^\ast P_0,` and :math:`P^\ast` is
    in general not a trimming operator.  (See :py:class:`~pyxu.operator.Pad`.)

    The same holds for gram/cogram operators.  Consider indeed the following order-1 backward finite-difference operator
    with zero-padding:

    .. code-block:: python3

       >>> S = Stencil(
       ...      arg_shape=(5,),
       ...      kernel=np.r_[-1, 1],
       ...      center=(0,),
       ...      mode='constant',
       ... )

       >>> S.gram().asarray()
       [[ 1  -1   0   0   0]
        [-1   2  -1   0   0]
        [ 0  -1   2  -1   0]
        [ 0   0  -1   2  -1]
        [ 0   0   0  -1   2]]

    We observe that the Gram differs from the order 2 centered finite-difference operator.  (Reduced-order derivative on
    one side.)

    Example
    -------

    * **Moving average of a 1D signal**

      Let :math:`x[n]` denote a 1D signal.  The weighted moving average

      .. math::

         y[n] = x[n-2] + 2 x[n-1] + 3 x[n]

      can be viewed as the output of the 3-point stencil of kernel :math:`k = [1, 2, 3]`.

      .. code-block:: python3

         import numpy as np
         from pyxu.operator import Stencil

         x = np.arange(10)  # [0  1  2  3  4  5  6  7  8  9]

         op = Stencil(
             arg_shape=x.shape,
             kernel=np.array([1, 2, 3]),
             center=(2,),  # k[2] applies on x[n]
         )

         y = op.apply(x)  # [0  3  8  14  20  26  32  38  44  50]


    * **Non-seperable image filtering**

      Let :math:`x[n, m]` denote a 2D image.  The blurred image

      .. math::

         y[n, m] = 2 x[n-1,m-1] + 3 x[n-1,m+1] + 4 x[n+1,m-1] + 5 x[n+1,m+1]

      can be viewed as the output of the 9-point stencil

      .. math::

         k =
         \left[
         \begin{array}{ccc}
            2 & 0 & 3 \\
            0 & 0 & 0 \\
            4 & 0 & 5
         \end{array}
         \right].

      .. code-block:: python3

         import numpy as np
         from pyxu.operator import Stencil

         x = np.arange(64).reshape(8, 8)  # square image
         # [[ 0   1   2   3   4   5   6   7]
         #  [ 8   9  10  11  12  13  14  15]
         #  [16  17  18  19  20  21  22  23]
         #  [24  25  26  27  28  29  30  31]
         #  [32  33  34  35  36  37  38  39]
         #  [40  41  42  43  44  45  46  47]
         #  [48  49  50  51  52  53  54  55]
         #  [56  57  58  59  60  61  62  63]]

         op = Stencil(
             arg_shape=x.shape,
             kernel=np.array(
                 [[2, 0, 3],
                  [0, 0, 0],
                  [4, 0, 5]]),
             center=(1, 1),  # k[1, 1] applies on x[n, m]
         )

         y = op.apply(x.reshape(-1)).reshape(8, 8)
         # [[ 45   82   91  100  109  118  127   56 ]
         #  [ 88  160  174  188  202  216  230  100 ]
         #  [152  272  286  300  314  328  342  148 ]
         #  [216  384  398  412  426  440  454  196 ]
         #  [280  496  510  524  538  552  566  244 ]
         #  [344  608  622  636  650  664  678  292 ]
         #  [408  720  734  748  762  776  790  340 ]
         #  [147  246  251  256  261  266  271  108 ]]

    * **Seperable image filtering**

      Let :math:`x[n, m]` denote a 2D image.  The warped image

      .. math::

         \begin{align*}
             y[n, m] = & +  4 x[n-1,m-1] +  5 x[n-1,m] +  6 x[n-1,m+1] \\
                       & +  8 x[n  ,m-1] + 10 x[n  ,m] + 12 x[n  ,m+1] \\
                       & + 12 x[n+1,m-1] + 15 x[n+1,m] + 18 x[n+1,m+1]
         \end{align*}

      can be viewed as the output of the 9-point stencil

      .. math::

         k_{2D} =
         \left[
         \begin{array}{ccc}
             4 &  5 &  6 \\
             8 & 10 & 12 \\
            12 & 15 & 18 \\
         \end{array}
         \right].

      Notice however that :math:`y[n, m]` can be implemented more efficiently by factoring the 9-point stencil as a
      cascade of two 3-point stencils:

      .. math::

         k_{2D}
         = k_{1} k_{2}^{T}
         = \left[
         \begin{array}{c}
            1 \\ 2 \\ 3
         \end{array}
         \right]
         \left[
         \begin{array}{c}
            4 & 5 & 6
         \end{array}
         \right].

      Seperable stencils are supported and should be preferred when applicable.

      .. code-block:: python3

         import numpy as np
         from pyxu.operator import Stencil

         x = np.arange(64).reshape(8, 8)  # square image
         # [[ 0   1   2   3   4   5   6   7]
         #  [ 8   9  10  11  12  13  14  15]
         #  [16  17  18  19  20  21  22  23]
         #  [24  25  26  27  28  29  30  31]
         #  [32  33  34  35  36  37  38  39]
         #  [40  41  42  43  44  45  46  47]
         #  [48  49  50  51  52  53  54  55]
         #  [56  57  58  59  60  61  62  63]]

         op_2D = Stencil(  # using non-seperable kernel
             arg_shape=x.shape,
             kernel=np.array(
                 [[ 4,  5,  6],
                  [ 8, 10, 12],
                  [12, 15, 18]]),
             center=(1, 1),  # k[1, 1] applies on x[n, m]
         )
         op_sep = Stencil(  # using seperable kernels
             arg_shape=x.shape,
             kernel=[
                 np.array([1, 2, 3]),  # k1: stencil along 1st axis
                 np.array([4, 5, 6]),  # k2: stencil along 2nd axis
             ],
             center=(1, 1),  # k1[1] * k2[1] applies on x[n, m]
         )

         y_2D = op_2D.apply(x.reshape(-1)).reshape(8, 8)
         y_sep = op_sep.apply(x.reshape(-1)).reshape(8, 8)  # np.allclose(y_2D, y_sep) -> True
         # [[ 294   445   520   595   670   745   820   511 ]
         #  [ 740  1062  1152  1242  1332  1422  1512   930 ]
         #  [1268  1782  1872  1962  2052  2142  2232  1362 ]
         #  [1796  2502  2592  2682  2772  2862  2952  1794 ]
         #  [2324  3222  3312  3402  3492  3582  3672  2226 ]
         #  [2852  3942  4032  4122  4212  4302  4392  2658 ]
         #  [3380  4662  4752  4842  4932  5022  5112  3090 ]
         #  [1778  2451  2496  2541  2586  2631  2676  1617 ]]

    See Also
    --------
    :py:class:`~pyxu.operator.Convolve`,
    :py:class:`~pyxu.operator._Stencil`

    """

    KernelSpec = typ.Union[
        pxt.NDArray,  # (k1, ..., kD) non-seperable kernel
        cabc.Sequence[pxt.NDArray],  # [(k1,), ..., (kD,)] seperable kernels
    ]

    def __init__(
        self,
        arg_shape: pxt.NDArrayShape,
        kernel: KernelSpec,
        center: _Stencil.IndexSpec,
        mode: Pad.ModeSpec = "constant",
        enable_warnings: bool = True,
    ):
        r"""
        Parameters
        ----------
        arg_shape: NDArrayShape
            Shape of the rank-:math:`D` input array.
        kernel: ~pyxu.operator.Stencil.KernelSpec
            Stencil coefficients.  Two forms are accepted:

            * NDArray of rank-:math:`D`: denotes a non-seperable stencil.
            * tuple[NDArray_1, ..., NDArray_D]: a sequence of 1D stencils such that dimension[k] is filtered by stencil
              `kernel[k]`, that is:

              .. math::

                 k = k_1\otimes \cdots\otimes k_D,

              or in Python: ``k = functools.reduce(numpy.multiply.outer, kernel)``.

        center: ~pyxu.operator._Stencil.IndexSpec
            (i_1, ..., i_D) index of the stencil's center.

            `center` defines how a kernel is overlaid on inputs to produce outputs.

        mode: str, :py:class:`list` ( str )
            Boundary conditions.  Multiple forms are accepted:

            * str: unique mode shared amongst dimensions.
              Must be one of:

              * 'constant' (zero-padding)
              * 'wrap'
              * 'reflect'
              * 'symmetric'
              * 'edge'
            * tuple[str, ...]: dimension[k] uses `mode[k]` as boundary condition.

            (See :py:func:`numpy.pad` for details.)
        enable_warnings: bool
            If ``True``, emit a warning in case of precision mis-match issues.
        """
        arg_shape, _kernel, _center, _mode = self._canonical_repr(arg_shape, kernel, center, mode)
        codim = dim = np.prod(arg_shape)
        super().__init__(shape=(codim, dim))

        # Pad/Trim operators
        pad_width = self._compute_pad_width(_kernel, _center, _mode)
        self._pad = Pad(
            arg_shape=arg_shape,
            pad_width=pad_width,
            mode=_mode,
        )
        self._trim = Trim(
            arg_shape=self._pad._pad_shape,
            trim_width=pad_width,
        )

        # Stencil operators used in .apply()
        self._st_fw = [None] * len(_kernel)
        for i, (k_fw, c_fw) in enumerate(zip(_kernel, _center)):
            self._st_fw[i] = _Stencil.init(kernel=k_fw, center=c_fw)

        # Stencil operators used in .adjoint()
        self._st_bw = [None] * len(_kernel)
        _kernel, _center = self._bw_equivalent(_kernel, _center)
        for i, (k_bw, c_bw) in enumerate(zip(_kernel, _center)):
            self._st_bw[i] = _Stencil.init(kernel=k_bw, center=c_bw)

        self._dispatch_params = dict()  # Extra kwargs passed to _Stencil.apply()
        self._dtype = _kernel[0].dtype  # useful constant
        self._enable_warnings = bool(enable_warnings)

        # We know a crude Lipschitz bound by default. Since computing it takes (code) space,
        # the estimate is computed as a special case of estimate_lipschitz()
        self.lipschitz = self.estimate_lipschitz(__rule=True)

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        x = self._pad.apply(arr)
        x = x.reshape(-1, *self._pad._pad_shape)

        y = self._stencil_chain(self._cast_warn(x), self._st_fw)
        y = y.reshape(*arr.shape[:-1], -1)

        out = self._trim.apply(y)
        return out

    @pxrt.enforce_precision(i="arr")
    def adjoint(self, arr: pxt.NDArray) -> pxt.NDArray:
        x = self._trim.adjoint(arr)
        x = x.reshape(-1, *self._pad._pad_shape)

        y = self._stencil_chain(self._cast_warn(x), self._st_bw)
        y = y.reshape(*arr.shape[:-1], -1)

        out = self._pad.adjoint(y)
        return out

    def configure_dispatcher(self, **kwargs):
        """
        (Only applies if `kernel` is a CuPy array.)

        Configure stencil Dispatcher.

        See :py:meth:`~pyxu.operator._Stencil.apply` for accepted options.

        Example
        -------
        .. code-block:: python3

           import cupy as cp
           from pyxu.operator import Stencil

           x = cp.arange(10)

           op = Stencil(
               arg_shape=x.shape,
               kernel=np.array([1, 2, 3]),
               center=(1,),
           )

           y = op.apply(x)  # uses default threadsperblock/blockspergrid values

           op.configure_dispatcher(
               threadsperblock=50,
               blockspergrid=3,
           )
           y = op.apply(x)  # supplied values used instead
        """
        for k, v in kwargs.items():
            self._dispatch_params.update(k=v)

    @staticmethod
    def _canonical_repr(arg_shape, kernel, center, mode):
        # Create canonical representations
        #   * `arg_shape`: tuple[int]
        #   * `_kernel`: list[ndarray[float], ...]
        #   * `_center`: list[ndarray[int], ...]
        #   * `_mode`: list[str, ...]
        if not isinstance(arg_shape, cabc.Sequence):
            arg_shape = (arg_shape,)
        arg_shape = tuple(arg_shape)

        N = len(arg_shape)
        assert len(center) == N

        kernel = pxu.compute(kernel, traverse=True)
        try:
            # array input -> non-seperable filter
            pxu.get_array_module(kernel)
            assert kernel.ndim == N

            _kernel = [pxrt.coerce(kernel)]
            _center = [np.array(center, dtype=int)]
        except Exception:
            # sequence input -> seperable filter(s)
            assert len(kernel) == N  # one filter per dimension

            _kernel = [None] * N
            for i in range(N):
                sh = [1] * N
                sh[i] = -1
                _kernel[i] = pxrt.coerce(kernel[i]).reshape(sh)

            _center = np.zeros((N, N), dtype=int)
            _center[np.diag_indices(N)] = center

        _mode = Pad(  # get `mode` in canonical form
            (3,) * _kernel[0].ndim,
            pad_width=1,
            mode=mode,
        )._mode

        return arg_shape, _kernel, _center, _mode

    @staticmethod
    def _compute_pad_width(_kernel, _center, _mode) -> Pad.WidthSpec:
        N = _kernel[0].ndim
        pad_width = [None] * N
        for i in range(N):
            if len(_kernel) == 1:  # non-seperable filter
                c = _center[0][i]
                n = _kernel[0].shape[i]
            else:  # seperable filter(s)
                c = _center[i][i]
                n = _kernel[i].size

            # 1. Pad/Trim operators are shared amongst [apply,adjoint]():
            #    lhs/rhs are thus padded equally.
            # 2. When mode != "constant", pad width must match kernel dimensions to retain border
            #    effects.
            if _mode[i] == "constant":
                p = max(c, n - c - 1)
            else:  # anything else supported by Pad()
                p = n - 1
            pad_width[i] = (p, p)
        return tuple(pad_width)

    @staticmethod
    def _bw_equivalent(_kernel, _center):
        # Transform FW kernel/center specification to BW variant.
        k_bw = [np.flip(k_fw) for k_fw in _kernel]

        if len(_kernel) == 1:  # non-seperable filter
            c_bw = [(_kernel[0].shape - _center[0]) - 1]
        else:  # seperable filter(s)
            N = _kernel[0].ndim
            c_bw = np.zeros((N, N), dtype=int)
            for i in range(N):
                c_bw[i, i] = _kernel[i].shape[i] - _center[i][i] - 1

        return k_bw, c_bw

    def _stencil_chain_DASK(self, x: pxt.NDArray, stencils: list) -> pxt.NDArray:
        # Apply sequence of stencils to `x`.
        #
        # x: (N_stack, N_1, ..., N_D)
        # y: (N_stack, N_1, ..., N_D)
        #
        # [2022.12.28, Sepand]
        #   For some unknown reason, .map_overlap() gives incorrect results when inputs to
        #   .map_overlap() contain stacking dimensions.
        #   Workaround: call .map_overlap() on each stack-dim seperately, then re-assemble.

        # _compute_pad_width(): LHS/RHS padded equally, so choose either one
        depth = [lhs for (lhs, rhs) in self._pad._pad_width]
        y = [
            _.map_overlap(
                self._stencil_chain,
                depth=depth,
                boundary=0,
                trim=True,
                dtype=x.dtype,
                meta=x._meta,
                stencils=stencils,
            )
            for _ in x
        ]

        xp = pxu.get_array_module(x)
        y = xp.stack(y, axis=0)
        return y

    @pxu.redirect("x", DASK=_stencil_chain_DASK)
    def _stencil_chain(self, x: pxt.NDArray, stencils: list) -> pxt.NDArray:
        # Apply sequence of stencils to `x`.
        # (For Dask inputs, see _stencil_chain_DASK().)
        #
        # x: (N_stack, N_1, ..., N_D)
        # y: (N_stack, N_1, ..., N_D)
        if len(stencils) == 1:
            x, y = x, x.copy()
        else:
            # [2023.04.17, Sepand]
            # In-place updates of `x` breaks thread-safety of Stencil().
            # This is problematic if Stencil() is used with DASK inputs.
            x, y = x.copy(), x.copy()

        for st in stencils:
            st.apply(x, y, **self._dispatch_params)
            x, y = y, x
        y = x
        return y

    def _cast_warn(self, arr: pxt.NDArray) -> pxt.NDArray:
        if arr.dtype == self._dtype:
            out = arr
        else:
            if self._enable_warnings:
                msg = "Computation may not be performed at the requested precision."
                warnings.warn(msg, pxw.PrecisionWarning)
            out = arr.astype(dtype=self._dtype)
        return out

    def estimate_lipschitz(self, **kwargs) -> pxt.Real:
        no_eval = "__rule" in kwargs
        if no_eval:
            # Analytic upper bound from Young's convolution inequality:
            #     \norm{x \ast h}{2} \le \norm{x}{2}\norm{h}{1}
            #
            # -> L \le \norm{h}{1}
            kernels = [st._kernel for st in self._st_fw]
            kernel = functools.reduce(operator.mul, kernels, 1)
            L_st = np.linalg.norm(pxu.to_NUMPY(kernel).reshape(-1), ord=1)

            L_pad = self._pad.lipschitz
            L_trim = self._trim.lipschitz

            L = L_trim * L_st * L_pad  # upper bound
        else:
            L = super().estimate_lipschitz(**kwargs)
        return L

    def to_sciop(self, **kwargs):
        # Stencil.apply/adjoint() prefer precision provided at init-time.
        kwargs.update(dtype=self._dtype)
        op = pxa.SquareOp.to_sciop(self, **kwargs)
        return op

    def asarray(self, **kwargs) -> pxt.NDArray:
        # Stencil.apply() prefers precision provided at init-time.
        xp = pxu.get_array_module(self._st_fw[0]._kernel)
        _A = super().asarray(xp=xp, dtype=self._dtype)

        xp = kwargs.get("xp", pxd.NDArrayInfo.NUMPY.module())
        dtype = kwargs.get("dtype", pxrt.getPrecision().value)
        A = xp.array(pxu.to_NUMPY(_A), dtype=dtype)
        return A

    @pxrt.enforce_precision()
    def trace(self, **kwargs) -> pxt.Real:
        if all(m == "constant" for m in self._pad._mode):
            # tr = (kernel center coefficient) * N
            tr = functools.reduce(
                operator.mul,
                [st._kernel[tuple(st._center)] for st in self._st_fw],
                1,
            )
            tr *= self.dim
        else:
            # Standard algorithm, with computations restricted to precision supported by
            # Stencil.apply().
            kwargs.update(dtype=self._dtype)
            tr = super().trace(**kwargs)
        return float(tr)

    @property
    def kernel(self) -> KernelSpec:
        r"""
        Stencil kernel coefficients.

        Returns
        -------
        kern: ~pyxu.operator.Stencil.KernelSpec
            Stencil coefficients.

            If the kernel is non-seperable, a single array is returned.
            Otherwise :math:`D` arrays are returned, one per axis.
        """
        if len(self._st_fw) == 1:
            kern = self._st_fw[0]._kernel
        else:
            kern = [st._kernel for st in self._st_fw]
        return kern

    @property
    def center(self) -> _Stencil.IndexSpec:
        """
        Stencil central position.

        Returns
        -------
        ctr: ~pyxu.operator._Stencil.IndexSpec
            Stencil central position.
        """
        if len(self._st_fw) == 1:
            ctr = self._st_fw[0]._center
        else:
            ctr = [st._center[d] for (d, st) in enumerate(self._st_fw)]
        return tuple(ctr)

    @property
    def relative_indices(self) -> cabc.Sequence[pxt.NDArray]:
        r"""
        Relative indices of the stencil.

        Returns
        -------
        r_idx: :py:class:`list` ( :py:attr:`~pyxu.info.ptype.NDArray` )
            Relative indices of the stencil's kernel in each dimension.

        Example
        -------
        .. code-block:: python3

           S = Stencil(
               arg_shape=(5, 6, 9),
               kernel=[
                   np.r_[1, -1],
                   np.r_[3, 2, 1],
                   np.r_[2, -1, 3, 1],
               ],
               center=(1, 0, 3),
           )
           S.relative_indices  # [array([-1,  0]), array([0, 1, 2]), array([-3, -2, -1,  0])]
        """
        if len(self._st_fw) == 1:
            r_idx = [np.arange(s) - c for c, s in zip(self.center, self.kernel.shape)]
        else:
            r_idx = [np.arange(k.size) - c for c, k in zip(self.center, self.kernel)]
        return r_idx

    def visualize(self) -> str:
        r"""
        Show the :math:`D`-dimensional stencil kernel.

        The stencil's center is identified by surrounding parentheses.

        Example
        -------
        .. code-block:: python3

           from pyxu.operator import Stencil

           S = Stencil(
               arg_shape=(5, 6),
               kernel=[
                   np.r_[3, 2, 1],
                   np.r_[2, -1, 3, 1],
               ],
               center=(1, 2),
           )
           print(S.visualize())  # [[6.0 -3.0  9.0  3.0]
                                 #  [4.0 -2.0 (6.0) 2.0]
                                 #  [2.0 -1.0  3.0 1.0]]
        """
        kernels = [st._kernel for st in self._st_fw]
        kernel = functools.reduce(operator.mul, kernels, 1)

        kernel = pxu.to_NUMPY(kernel).astype(str)
        kernel[self.center] = "(" + kernel[self.center] + ")"

        kern = np.array2string(kernel).replace("'", "")
        return kern


Correlate = Stencil  #: Alias of :py:class:`~pyxu.operator.Stencil`.


class Convolve(Stencil):
    r"""
    Multi-dimensional JIT-compiled convolution.

    Inputs are convolved with the given kernel.

    Notes
    -----
    Given a :math:`D`-dimensional array :math:`x\in\mathbb{R}^{N_1\times\cdots\times N_D}` and kernel
    :math:`k\in\mathbb{R}^{K_1\times\cdots\times K_D}` with center :math:`(c_1, \ldots, c_D)`, the output of the
    convolution operator is an array :math:`y\in\mathbb{R}^{N_1\times\cdots\times N_D}` given by:

    .. math::

       y[i_{1},\ldots,i_{D}]
       =
       \sum_{q_{1},\ldots,q_{D}=0}^{K_{1},\ldots,K_{D}}
       x[i_{1} - q_{1} + c_{1},\ldots,i_{D} - q_{D} + c_{D}]
       \,\cdot\,
       k[q_{1},\ldots,q_{D}].

    The convolution is implemented via :py:class:`~pyxu.operator.Stencil`.  To do so, the convolution kernel is
    transformed to the equivalent correlation kernel:

    .. math::

       y[i_{1},\ldots,i_{D}]
       =
       \sum_{q_{1},\ldots,q_{D}=0}^{K_{1},\ldots,K_{D}}
       &x[i_{1} + q_{1} - (K_{1} - c_{1}),\ldots,i_{D} + q_{D} - (K_{D} - c_{D})] \\
       &\cdot\,
       k[K_{1}-q_{1},\ldots,K_{D}-q_{D}].

    This corresponds to a correlation with a flipped kernel and center.

    Examples
    --------
    .. code-block:: python3

       import numpy as np
       from scipy.ndimage import convolve
       from pyxu.operator import Convolve

       x = np.array([
            [1, 2, 0, 0],
            [5, 3, 0, 4],
            [0, 0, 0, 7],
            [9, 3, 0, 0],
       ])
       k = np.array([
            [1, 1, 1],
            [1, 1, 0],
            [1, 0, 0],
       ])
       op = Convolve(
           arg_shape=x.shape,
           kernel=k,
           center=(1, 1),
           mode="constant",
       )

       y_op = op.apply(x.reshape(-1)).reshape(4, 4)
       y_sp = convolve(x, k, mode="constant", origin=0)  # np.allclose(y_op, y_sp) -> True
       # [[11  10   7   4],
       #  [10   3  11  11],
       #  [15  12  14   7],
       #  [12   3   7   0]]

    See Also
    --------
    :py:class:`~pyxu.operator.Stencil`
    """

    def __init__(
        self,
        arg_shape: pxt.NDArrayShape,
        kernel: Stencil.KernelSpec,
        center: _Stencil.IndexSpec,
        mode: Pad.ModeSpec = "constant",
        enable_warnings: bool = True,
    ):
        r"""
        See :py:meth:`~pyxu.operator.Stencil.__init__` for a description of the arguments.
        """
        super().__init__(
            arg_shape=arg_shape,
            kernel=kernel,
            center=center,
            mode=mode,
            enable_warnings=enable_warnings,
        )

        # flip FW/BW kernels (& centers)
        self._st_fw, self._st_bw = self._st_bw, self._st_fw



import collections.abc as cabc
import functools

import dask

import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.util.misc as pxm

__all__ = [
    "compute",
    "get_array_module",
    "redirect",
    "to_NUMPY",
]


def get_array_module(x, fallback: pxt.ArrayModule = None) -> pxt.ArrayModule:
    """
    Get the array namespace corresponding to a given object.

    Parameters
    ----------
    x: object
        Any object compatible with the interface of NumPy arrays.
    fallback: ArrayModule
        Fallback module if `x` is not a NumPy-like array.  Default behaviour: raise error if fallback used.

    Returns
    -------
    namespace: ArrayModule
        The namespace to use to manipulate `x`, or `fallback` if provided.
    """

    def infer_api(y):
        try:
            return pxd.NDArrayInfo.from_obj(y).module()
        except ValueError:
            return None

    if (xp := infer_api(x)) is not None:
        return xp
    elif fallback is not None:
        return fallback
    else:
        raise ValueError(f"Could not infer array module for {type(x)}.")


def redirect(
    i: pxt.VarName,
    **kwargs: cabc.Mapping[str, cabc.Callable],
) -> cabc.Callable:
    """
    Change codepath for supplied array backends.

    Some functions/methods cannot be written in module-agnostic fashion. The action of this decorator is summarized
    below:

    * Analyze an array-valued parameter (`x`) of the wrapped function/method (`f`).
    * If `x` lies in one of the supplied array namespaces: re-route execution to the specified function.
    * If `x` lies in none of the supplied array namespaces: execute `f`.

    Parameters
    ----------
    i: VarName
        name of the array-like variable in `f` to base dispatch on.
    kwargs: ~collections.abc.Mapping

        * key[:py:class:`str`]: array backend short-name as defined in :py:class:`~pyxu.info.deps.NDArrayInfo`.
        * value[:py:class:`collections.abc.Callable`]: function/method to dispatch to.

    Notes
    -----
    Auto-dispatch via :py:func:`redirect` assumes the dispatcher/dispatchee have the same parameterization, i.e.:

    * if `f` is a function -> dispatch possible to another callable with identical signature (i.e., function or
      staticmethod)
    * if `f` is a staticmethod -> dispatch possible to another callable with identical signature (i.e., function or
      staticmethod)
    * if `f` is an instance-method -> dispatch to another instance-method of the class with identical signature.

    Example
    -------
    .. code-block:: python3

       def f(x, y): return "f"

       @redirect('x', NUMPY=f)    # if 'x' is of type NDArrayInfo.NUMPY, i.e. has
       def g(x, y): return "g"    # short-name 'NUMPY' -> reroute execution to `f`

       x1 = np.arange(5)
       x2 = da.array(x1)
       y = 1
       g(x1, y), g(x2, y)  # 'f', 'g'
    """

    def decorator(func: cabc.Callable) -> cabc.Callable:
        @functools.wraps(func)
        def wrapper(*ARGS, **KWARGS):
            try:
                func_args = pxm.parse_params(func, *ARGS, **KWARGS)
            except Exception as e:
                error_msg = f"Could not parameterize {func}()."
                raise ValueError(error_msg) from e

            if i not in func_args:
                error_msg = f"Parameter[{i}] not part of {func.__qualname__}() parameter list."
                raise ValueError(error_msg)

            ndi = pxd.NDArrayInfo.from_obj(func_args[i])
            if (alt_func := kwargs.get(ndi.name)) is not None:
                out = alt_func(**func_args)
            else:
                out = func(**func_args)

            return out

        return wrapper

    return decorator


def compute(*args, mode: str = "compute", **kwargs):
    r"""
    Force computation of Dask collections.

    Parameters
    ----------
    \*args: object, list
        Any number of objects.  If it is a dask object, it is evaluated and the result is returned.  Non-dask arguments
        are passed through unchanged.  Python collections are traversed to find/evaluate dask objects within.  (Use
        `traverse` =False to disable this behavior.)
    mode: str
        Dask evaluation strategy: compute or persist.
    \*\*kwargs: dict
        Extra keyword parameters forwarded to :py:func:`dask.compute` or :py:func:`dask.persist`.

    Returns
    -------
    \*cargs: object, list
        Evaluated objects. Non-dask arguments are passed through unchanged.
    """
    try:
        func = dict(compute=dask.compute, persist=dask.persist)[mode.lower()]
    except Exception:
        raise ValueError(f"mode: expected compute/persist, got {mode}.")

    cargs = func(*args, **kwargs)
    if len(args) == 1:
        cargs = cargs[0]
    return cargs


def to_NUMPY(x: pxt.NDArray) -> pxt.NDArray:
    """
    Convert an array from a specific backend to NUMPY.

    Parameters
    ----------
    x: NDArray
        Array to be converted.

    Returns
    -------
    y: NDArray
        Array with NumPy backend.

    Notes
    -----
    This function is a no-op if the array is already a NumPy array.
    """
    N = pxd.NDArrayInfo
    ndi = N.from_obj(x)
    if ndi == N.NUMPY:
        y = x
    elif ndi == N.DASK:
        y = compute(x)
    elif ndi == N.CUPY:
        y = x.get()
    else:
        msg = f"Dev-action required: define behaviour for {ndi}."
        raise ValueError(msg)
    return y



import warnings

import scipy.sparse.linalg as spsl

import pyxu.abc as pxa
import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.info.warning as pxw
import pyxu.operator.interop.source as px_src
import pyxu.runtime as pxrt
import pyxu.util as pxu

__all__ = [
    "from_sciop",
]


def from_sciop(cls: pxt.OpC, sp_op: spsl.LinearOperator) -> pxt.OpT:
    r"""
    Wrap a :py:class:`~scipy.sparse.linalg.LinearOperator` as a :py:class:`~pyxu.abc.LinOp` (or sub-class thereof).

    Parameters
    ----------
    sp_op: ~scipy.sparse.linalg.LinearOperator
        (N, M) Linear CPU/GPU operator compliant with SciPy's interface.

    Returns
    -------
    op: OpT
        (N, M) Pyxu-compliant linear operator.
    """
    assert cls.has(pxa.Property.LINEAR)

    if sp_op.dtype not in [_.value for _ in pxrt.Width]:
        warnings.warn(
            "Computation may not be performed at the requested precision.",
            pxw.PrecisionWarning,
        )

    # [r]matmat only accepts 2D inputs -> reshape apply|adjoint inputs as needed.

    @pxrt.enforce_precision(i="arr")
    def op_apply(_, arr: pxt.NDArray) -> pxt.NDArray:
        sh = arr.shape[:-1]
        arr = arr.reshape(-1, _.dim)
        out = _._sp_op.matmat(arr.T).T
        out = out.reshape(*sh, _.codim)
        return out

    @pxrt.enforce_precision(i="arr")
    def op_adjoint(_, arr: pxt.NDArray) -> pxt.NDArray:
        sh = arr.shape[:-1]
        arr = arr.reshape(-1, _.codim)
        out = _._sp_op.rmatmat(arr.T).T
        out = out.reshape(*sh, _.dim)
        return out

    def op_asarray(_, **kwargs) -> pxt.NDArray:
        # Determine XP-module accepted by sci_op, then compute array-representation.
        for ndi in [
            pxd.NDArrayInfo.NUMPY,
            pxd.NDArrayInfo.CUPY,
        ]:
            try:
                cls = _.__class__
                _A = cls.asarray(_, xp=ndi.module(), dtype=_._sp_op.dtype)
                break
            except Exception:
                pass

        # Cast to user specs.
        xp = kwargs.get("xp", pxd.NDArrayInfo.NUMPY.module())
        dtype = kwargs.get("dtype", pxrt.getPrecision().value)
        A = xp.array(pxu.to_NUMPY(_A), dtype=dtype)
        return A

    def op_expr(_) -> tuple:
        return ("from_sciop", _._sp_op)

    op = px_src.from_source(
        cls=cls,
        shape=sp_op.shape,
        apply=op_apply,
        adjoint=op_adjoint,
        asarray=op_asarray,
        _expr=op_expr,
    )
    op._sp_op = sp_op

    return op



import itertools
import typing as typ

import numpy as np
import pytest

import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.operator as pxo
import pyxu.runtime as pxrt
import pyxu.util as pxu
import pyxu_tests.operator.conftest as conftest


class TestFFT(conftest.LinOpT):
    # Internal Helpers ----------------------------------------------------------------------------
    @staticmethod
    def spec_data() -> (
        list[
            tuple[
                pxd.NDArrayInfo,
                pxrt.Width,
                typ.Any,  # (arg_shape, axes), user-provided to FFT.__init__()
                typ.Any,  # (arg_shape, axes), ground-truth it corresponds to. (Canonical form.)
                dict,  # kwargs
            ]
        ]
    ):
        N = pxd.NDArrayInfo
        W = pxrt.Width
        data = []

        # (arg_shape, axes) configs to test
        # * `spec[k][0]` corresponds to raw inputs users provide to FFT().
        # * `spec[k][1]` corresponds to their ground-truth canonical parameterization.
        spec = [
            # 1D transforms ---------------------------------------------------
            (
                (5, None),
                ((5,), (0,)),
            ),
            (
                ((5,), None),
                ((5,), (0,)),
            ),
            (
                (5, 0),
                ((5,), (0,)),
            ),
            # ND transforms ---------------------------------------------------
            (
                ((5, 3, 4), None),
                ((5, 3, 4), (0, 1, 2)),
            ),
            (
                ((5, 3, 4), 0),
                ((5, 3, 4), (0,)),
            ),
            (
                ((5, 3, 4), 1),
                ((5, 3, 4), (1,)),
            ),
            (
                ((5, 3, 4), 2),
                ((5, 3, 4), (2,)),
            ),
            (
                ((5, 3, 4), (0, 1)),  # transform axes sequential
                ((5, 3, 4), (0, 1)),
            ),
            (
                ((5, 3, 4), (0, 2)),  # transform axes non-sequential
                ((5, 3, 4), (0, 2)),
            ),
        ]

        # Test all backend/width combos, with no kwargs -----------------------
        for ndi, width, (init_spec, canonical_spec) in itertools.product(N, W, spec):
            data.append(
                (
                    ndi,
                    width,
                    init_spec,
                    canonical_spec,
                    dict(),  # no kwargs
                )
            )

        # A specific NUMPY/DASK spec, with all kwargs combos ------------------
        kwargs = [
            dict(
                workers=workers,
                auto_align_input=auto_align_input,
                auto_contiguous=auto_contiguous,
            )
            for (workers, auto_align_input, auto_contiguous) in itertools.product(
                [1, 2],  # workers
                [True, False],  # auto_align_input
                [True, False],  # auto_contiguous
            )
        ]
        for ndi, width, (init_spec, canonical_spec), _kwargs in itertools.product(
            [N.NUMPY, N.DASK],
            W,
            [spec[5]],  # ND input, mono-axes transform
            kwargs,
        ):
            data.append(
                (
                    ndi,
                    width,
                    init_spec,
                    canonical_spec,
                    _kwargs,
                )
            )

        return data

    @pytest.fixture(params=spec_data())
    def _spec(self, request):
        return request.param

    @pytest.fixture
    def arg_shape(self, _spec):  # canonical representation
        arg_shape = _spec[3][0]
        return arg_shape

    @pytest.fixture
    def axes(self, _spec):  # canonical representation
        axes = _spec[3][1]
        return axes

    @pytest.fixture(params=[False, True])
    def transform_real(self, request) -> bool:
        return request.param

    @pytest.fixture
    def spec(self, _spec, transform_real) -> tuple[pxt.OpT, pxd.NDArrayInfo, pxrt.Width]:
        ndi, width, (arg_shape, axes), _, kwargs = _spec  # user-provided (arg_shape, axes).

        op = pxo.FFT(
            arg_shape=arg_shape,
            axes=axes,
            real=transform_real,
            **kwargs,
        )
        return op, ndi, width

    @pytest.fixture
    def data_shape(self, arg_shape, transform_real) -> pxt.OpShape:
        codim = dim = 2 * np.prod(arg_shape)
        if transform_real:
            dim /= 2
        return (codim, dim)

    @pytest.fixture
    def data_apply(self, arg_shape, axes, transform_real) -> conftest.DataLike:
        arr_r, arr_c = self._random_array((2, *arg_shape), seed=26)  # fixed seed for reproducibility
        if transform_real:
            arr = arr_r
            arr_gt = arr.reshape(-1)
        else:
            arr = arr_r + 1j * arr_c
            arr_gt = pxu.view_as_real(arr.reshape(-1))

        out = np.fft.fftn(arr, axes=axes, norm="backward")
        out_gt = pxu.view_as_real(out.reshape(-1))

        return dict(
            in_=dict(arr=arr_gt),
            out=out_gt,
        )

    # Overridden Tests --------------------------------------------------------
    def test_value_to_sciop(self, _op_sciop, _data_to_sciop):
        if _data_to_sciop["mode"] in {"matmat", "rmatmat"}:
            pytest.xfail(reason="Last axis is non-contiguous: apply/adjoint will fail.")
        else:
            super().test_value_to_sciop(_op_sciop, _data_to_sciop)

    def test_backend_to_sciop(self, _op_sciop, _data_to_sciop):
        if _data_to_sciop["mode"] in {"matmat", "rmatmat"}:
            pytest.xfail(reason="Last axis is non-contiguous: apply/adjoint will fail.")
        else:
            super().test_backend_to_sciop(_op_sciop, _data_to_sciop)

    def test_prec_to_sciop(self, _op_sciop, _data_to_sciop):
        if _data_to_sciop["mode"] in {"matmat", "rmatmat"}:
            pytest.xfail(reason="Last axis is non-contiguous: apply/adjoint will fail.")
        else:
            super().test_prec_to_sciop(_op_sciop, _data_to_sciop)



import itertools

import numpy as np
import pytest

import pyxu.info.deps as pxd
import pyxu.operator as pxo
import pyxu.runtime as pxrt
import pyxu_tests.operator.conftest as conftest


class TestL1Norm(conftest.ProxFuncT):
    @pytest.fixture(
        params=itertools.product(
            ((5, pxo.L1Norm(dim=5)),),  # dim, op
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def _spec(self, request):
        return request.param

    @pytest.fixture
    def spec(self, _spec):
        return _spec[0][1], _spec[1], _spec[2]

    @pytest.fixture
    def dim(self, _spec):
        return _spec[0][0]

    @pytest.fixture
    def data_shape(self, dim):
        return (1, dim)

    @pytest.fixture(
        params=[  # 2 evaluation points
            dict(
                in_=dict(arr=np.zeros((5,))),
                out=np.zeros((1,)),
            ),
            dict(
                in_=dict(arr=np.arange(-3, 2)),
                out=np.array([7]),
            ),
        ]
    )
    def data_apply(self, request):
        return request.param

    @pytest.fixture(
        params=[  # 2 evaluation points
            dict(
                in_=dict(
                    arr=np.zeros((5,)),
                    tau=1,
                ),
                out=np.zeros((5,)),
            ),
            dict(
                in_=dict(
                    arr=np.arange(-3, 2),
                    tau=1,
                ),
                out=np.array([-2, -1, 0, 0, 0]),
            ),
        ]
    )
    def data_prox(self, request):
        return request.param

    @pytest.fixture
    def data_math_lipschitz(self, dim):
        N_test, dim = 10, self._sanitize(dim, 3)
        return self._random_array((N_test, dim))


class TestL2Norm(conftest.ProxFuncT):
    @pytest.fixture(
        params=itertools.product(
            ((5, pxo.L2Norm(dim=5)),),  # dim, op
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def _spec(self, request):
        return request.param

    @pytest.fixture
    def spec(self, _spec):
        return _spec[0][1], _spec[1], _spec[2]

    @pytest.fixture
    def dim(self, _spec):
        return _spec[0][0]

    @pytest.fixture
    def data_shape(self, dim):
        return (1, dim)

    @pytest.fixture(
        params=[  # 2 evaluation points
            dict(
                in_=dict(arr=np.zeros((5,))),
                out=np.zeros((1,)),
            ),
            dict(
                in_=dict(arr=np.arange(-3, 2)),
                out=np.sqrt([15]),
            ),
        ]
    )
    def data_apply(self, request):
        return request.param

    @pytest.fixture(
        params=[  # 2 evaluation points
            dict(
                in_=dict(
                    arr=np.zeros((5,)),
                    tau=1,
                ),
                out=np.zeros((5,)),
            ),
            dict(
                in_=dict(
                    arr=np.arange(-3, 2),
                    tau=1,
                ),
                out=(1 - 1 / np.sqrt(15)) * np.arange(-3, 2),
            ),
        ]
    )
    def data_prox(self, request):
        return request.param

    @pytest.fixture
    def data_math_lipschitz(self, dim):
        N_test, dim = 10, self._sanitize(dim, 3)
        return self._random_array((N_test, dim))


class TestSquaredL2Norm(conftest.QuadraticFuncT):
    @pytest.fixture(
        params=itertools.product(
            ((7, pxo.SquaredL2Norm(dim=7)),),  # dim, op
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def _spec(self, request):
        return request.param

    @pytest.fixture
    def spec(self, _spec):
        return _spec[0][1], _spec[1], _spec[2]

    @pytest.fixture
    def dim(self, _spec):
        return _spec[0][0]

    @pytest.fixture
    def data_shape(self, dim):
        return (1, dim)

    @pytest.fixture(
        params=[  # 2 evaluation points
            dict(
                in_=dict(arr=np.zeros((7,))),
                out=np.zeros((1,)),
            ),
            dict(
                in_=dict(arr=np.arange(-3, 4)),
                out=np.array([28]),
            ),
        ]
    )
    def data_apply(self, request):
        return request.param

    @pytest.fixture(
        params=[  # 2 evaluation points
            dict(
                in_=dict(arr=np.zeros((7,))),
                out=np.zeros((7,)),
            ),
            dict(
                in_=dict(arr=np.arange(-3, 4)),
                out=2 * np.arange(-3, 4),
            ),
        ]
    )
    def data_grad(self, request):
        return request.param

    @pytest.fixture(
        params=[  # 2 evaluation points
            dict(
                in_=dict(
                    arr=np.zeros((7,)),
                    tau=1,
                ),
                out=np.zeros((7,)),
            ),
            dict(
                in_=dict(
                    arr=np.arange(-3, 4),
                    tau=1,
                ),
                out=np.arange(-3, 4) / 3,
            ),
        ]
    )
    def data_prox(self, request):
        return request.param


# We disable PerformanceWarnings due to prox_algo="sort" used on Dask inputs.
@pytest.mark.filterwarnings("ignore::pyxu.info.warning.PerformanceWarning")
class TestSquaredL1Norm(conftest.ProxFuncT):
    @pytest.fixture(
        params=itertools.product(
            (  # dim, op
                (5, pxo.SquaredL1Norm(dim=5, prox_algo="sort")),
                (5, pxo.SquaredL1Norm(dim=5, prox_algo="root")),
            ),
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def _spec(self, request):
        return request.param

    @pytest.fixture
    def spec(self, _spec):
        return _spec[0][1], _spec[1], _spec[2]

    @pytest.fixture
    def dim(self, _spec):
        return _spec[0][0]

    @pytest.fixture
    def data_shape(self, dim):
        return (1, dim)

    @pytest.fixture(
        params=[  # 2 evaluation points
            dict(
                in_=dict(arr=np.zeros((5,))),
                out=np.zeros((1,)),
            ),
            dict(
                in_=dict(arr=np.arange(-3, 2)),
                out=np.array([49]),
            ),
        ]
    )
    def data_apply(self, request):
        return request.param

    @pytest.fixture(
        params=[  # 2 evaluation points
            dict(
                in_=dict(
                    arr=np.zeros((5,)),
                    tau=1,
                ),
                out=np.zeros((5,)),
            ),
            dict(
                in_=dict(
                    arr=np.arange(-3, 2),
                    tau=1,
                ),
                out=np.array([-1, 0, 0, 0, 0]),
            ),
        ]
    )
    def data_prox(self, request):
        return request.param

    @pytest.fixture
    def data_math_lipschitz(self, dim):
        N_test, dim = 10, self._sanitize(dim, 3)
        return self._random_array((N_test, dim))


class TestLInfinityNorm(conftest.ProxFuncT):
    @pytest.fixture(
        params=itertools.product(
            ((5, pxo.LInfinityNorm(dim=5)),),  # dim, op
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def _spec(self, request):
        return request.param

    @pytest.fixture
    def spec(self, _spec):
        return _spec[0][1], _spec[1], _spec[2]

    @pytest.fixture
    def dim(self, _spec):
        return _spec[0][0]

    @pytest.fixture
    def data_shape(self, dim):
        return (1, dim)

    @pytest.fixture(
        params=[  # 2 evaluation points
            dict(
                in_=dict(arr=np.zeros((5,))),
                out=np.zeros((1,)),
            ),
            dict(
                in_=dict(arr=np.arange(-3, 2)),
                out=np.array([3]),
            ),
        ]
    )
    def data_apply(self, request):
        return request.param

    @pytest.fixture(
        params=[  # 2 evaluation points
            dict(
                in_=dict(
                    arr=np.zeros((5,)),
                    tau=1,
                ),
                out=np.zeros((5,)),
            ),
            dict(
                in_=dict(
                    arr=np.arange(-3, 2),
                    tau=1,
                ),
                out=np.array([-2, -2, -1, 0, 1]),
            ),
        ]
    )
    def data_prox(self, request):
        return request.param

    @pytest.fixture
    def data_math_lipschitz(self, dim):
        N_test, dim = 10, self._sanitize(dim, 3)
        return self._random_array((N_test, dim))


class TestL21Norm(conftest.ProxFuncT):
    @pytest.fixture(
        params=itertools.product(
            ((6, pxo.L21Norm(arg_shape=(2, 3))),),  # dim, op
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def _spec(self, request):
        return request.param

    @pytest.fixture
    def spec(self, _spec):
        return _spec[0][1], _spec[1], _spec[2]

    @pytest.fixture
    def dim(self, _spec):
        return _spec[0][0]

    @pytest.fixture
    def data_shape(self, dim):
        return (1, dim)

    @pytest.fixture(
        params=[  # 2 evaluation points
            dict(
                in_=dict(arr=np.zeros((6,))),
                out=np.zeros((1,)),
            ),
            dict(
                in_=dict(arr=np.array([1, 2, -3, 0, -2, -4])),
                out=np.array([6 + 2 * np.sqrt(2)]),
            ),
        ]
    )
    def data_apply(self, request):
        return request.param

    @pytest.fixture(
        params=[  # 2 evaluation points
            dict(
                in_=dict(
                    arr=np.zeros((6,)),
                    tau=1,
                ),
                out=np.zeros((6,)),
            ),
            dict(
                in_=dict(
                    arr=np.array([1, 2, -3, 0, -2, -4]),
                    tau=4,
                ),
                out=np.array([0, 0, -3 / 5, 0, 0, -4 / 5]),
            ),
        ]
    )
    def data_prox(self, request):
        return request.param

    @pytest.fixture
    def data_math_lipschitz(self, dim):
        N_test = 10
        return self._random_array((N_test, dim))


class TestPositiveL1Norm(conftest.ProxFuncT):
    @pytest.fixture(
        params=itertools.product(
            ((5, pxo.PositiveL1Norm(dim=5)),),  # dim, op
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def _spec(self, request):
        return request.param

    @pytest.fixture
    def spec(self, _spec):
        return _spec[0][1], _spec[1], _spec[2]

    @pytest.fixture
    def dim(self, _spec):
        return _spec[0][0]

    @pytest.fixture
    def data_shape(self, dim):
        return (1, dim)

    @pytest.fixture(
        params=[  # 3 evaluation points
            dict(
                in_=dict(arr=np.zeros((5,))),
                out=np.zeros((1,)),
            ),
            dict(
                in_=dict(arr=np.arange(-3, 2)),
                out=np.array([np.inf]),
            ),
            dict(
                in_=dict(arr=np.arange(5)),
                out=np.array([10]),
            ),
        ]
    )
    def data_apply(self, request):
        return request.param

    @pytest.fixture(
        params=[  # 3 evaluation points
            dict(
                in_=dict(
                    arr=np.zeros((5,)),
                    tau=1,
                ),
                out=np.zeros((5,)),
            ),
            dict(
                in_=dict(
                    arr=np.arange(-3, 2),
                    tau=1,
                ),
                out=np.array([0, 0, 0, 0, 0]),
            ),
            dict(
                in_=dict(
                    arr=np.arange(5),
                    tau=1,
                ),
                out=np.array([0, 0, 1, 2, 3]),
            ),
        ]
    )
    def data_prox(self, request):
        return request.param

    @pytest.fixture
    def data_math_lipschitz(self, dim):
        N_test, dim = 10, self._sanitize(dim, 3)
        return self._random_array((N_test, dim))
